import collections
import copy
import dataclasses
from enum import Enum
from typing import List, Dict, Iterator, Tuple

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackEnum, BitPackValue, BitPackDecoder
from randovania.game_description import data_reader
from randovania.game_description.resources.trick_resource_info import TrickResourceInfo
from randovania.games.game import RandovaniaGame
from randovania.games.prime import default_data
from randovania.interface_common.enum_lib import iterate_enum


class LayoutTrickLevel(BitPackEnum, Enum):
    DISABLED = "disabled"
    BEGINNER = "beginner"
    INTERMEDIATE = "intermediate"
    ADVANCED = "advanced"
    EXPERT = "expert"
    HYPERMODE = "hypermode"

    @classmethod
    def default(cls) -> "LayoutTrickLevel":
        return cls.DISABLED

    @classmethod
    def from_number(cls, number: int) -> "LayoutTrickLevel":
        return _TRICK_LEVEL_ORDER[number]

    @property
    def as_number(self) -> int:
        return _TRICK_LEVEL_ORDER.index(self)

    @property
    def long_name(self) -> str:
        return _PRETTY_TRICK_LEVEL_NAME[self]


_TRICK_LEVEL_ORDER: List[LayoutTrickLevel] = list(LayoutTrickLevel)
_PRETTY_TRICK_LEVEL_NAME = {
    LayoutTrickLevel.DISABLED: "Disabled",
    LayoutTrickLevel.BEGINNER: "Beginner",
    LayoutTrickLevel.INTERMEDIATE: "Intermediate",
    LayoutTrickLevel.ADVANCED: "Advanced",
    LayoutTrickLevel.EXPERT: "Expert",
    LayoutTrickLevel.HYPERMODE: "Hypermode",
}


def _all_tricks(game_data: dict):
    resource_database = data_reader.read_resource_database(game_data["resource_database"])
    return resource_database.trick


@dataclasses.dataclass(frozen=True)
class TrickLevelConfiguration(BitPackValue):
    minimal_logic: bool
    specific_levels: Dict[str, LayoutTrickLevel]
    game: RandovaniaGame

    def __post_init__(self):
        for trick, level in self.specific_levels.items():
            if not isinstance(level, LayoutTrickLevel) or level == LayoutTrickLevel.DISABLED:
                raise ValueError(f"Invalid level `{level}` for trick {trick}, "
                                 f"expected a LayoutTrickLevel that isn't NO_TRICKS")

    def bit_pack_encode(self, metadata) -> Iterator[Tuple[int, int]]:
        game_data = default_data.read_json_then_binary(self.game)[1]

        yield from bitpacking.encode_bool(self.minimal_logic)
        if self.minimal_logic:
            return

        encodable_levels = list(LayoutTrickLevel)
        encodable_levels.remove(LayoutTrickLevel.DISABLED)

        for trick in sorted(_all_tricks(game_data)):
            has_trick = self.has_specific_level_for_trick(trick)
            yield from bitpacking.encode_bool(has_trick)
            if has_trick:
                yield from bitpacking.pack_array_element(self.level_for_trick(trick), encodable_levels)

    @classmethod
    def bit_pack_unpack(cls, decoder: BitPackDecoder, metadata):
        game = metadata["reference"].game
        game_data = default_data.read_json_then_binary(game)[1]

        minimal_logic = bitpacking.decode_bool(decoder)
        specific_levels = {}

        if not minimal_logic:
            encodable_levels = list(LayoutTrickLevel)
            encodable_levels.remove(LayoutTrickLevel.DISABLED)

            for trick in sorted(_all_tricks(game_data)):
                if bitpacking.decode_bool(decoder):
                    specific_levels[trick.short_name] = decoder.decode_element(encodable_levels)

        return cls(minimal_logic, specific_levels, game)

    @property
    def pretty_description(self) -> str:
        if self.minimal_logic:
            return "Minimal Logic"

        trick_list = _all_tricks(default_data.read_json_then_binary(self.game)[1])
        difficulties = collections.defaultdict(int)
        for trick in trick_list:
            difficulties[self.level_for_trick(trick)] += 1

        if len(difficulties) == 1:
            for level in difficulties.keys():
                return f"All at {level.long_name}"

        descriptions = [
            f"{difficulties[level]} at {level.long_name}"
            for level in iterate_enum(LayoutTrickLevel)
            if difficulties[level] > 0
        ]
        return ", ".join(descriptions)

    @property
    def as_json(self) -> dict:
        specific_levels = {
            trick_short_name: level.value
            for trick_short_name, level in self.specific_levels.items()
        }

        return {
            "minimal_logic": self.minimal_logic,
            "specific_levels": {} if self.minimal_logic else specific_levels,
        }

    @classmethod
    def from_json(cls, value: dict, game: RandovaniaGame):
        minimal_logic = value["minimal_logic"]
        specific_levels = {
            trick_short_name: LayoutTrickLevel(level)
            for trick_short_name, level in value["specific_levels"].items()
            if level != LayoutTrickLevel.DISABLED.value
        }
        return cls(
            minimal_logic=minimal_logic,
            specific_levels={} if minimal_logic else specific_levels,
            game=game,
        )

    def has_specific_level_for_trick(self, trick: TrickResourceInfo) -> bool:
        return trick.short_name in self.specific_levels

    def level_for_trick(self, trick: TrickResourceInfo) -> LayoutTrickLevel:
        return self.specific_levels.get(trick.short_name, LayoutTrickLevel.DISABLED)

    def set_level_for_trick(self, trick: TrickResourceInfo, value: LayoutTrickLevel) -> "TrickLevelConfiguration":
        """
        Creates a new TrickLevelConfiguration with the given trick with a changed level
        :param trick:
        :param value:
        :return:
        """
        assert value is not None
        new_levels = copy.copy(self.specific_levels)

        if value != LayoutTrickLevel.DISABLED:
            new_levels[trick.short_name] = value
        elif trick.short_name in new_levels:
            del new_levels[trick.short_name]

        return dataclasses.replace(self, specific_levels=new_levels)

    def dangerous_settings(self) -> List[str]:
        if self.minimal_logic:
            return ["Minimal Logic"]
        return []



import dataclasses
from typing import Tuple, Iterator

from randovania.bitpacking import bitpacking
from randovania.bitpacking.bitpacking import BitPackDecoder
from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.item.major_item import MajorItem

ENERGY_TANK_MAXIMUM_COUNT = 16
DEFAULT_MAXIMUM_SHUFFLED = (4, 10, 99)


@dataclasses.dataclass(frozen=True)
class MajorItemState:
    include_copy_in_original_location: bool = False
    num_shuffled_pickups: int = 0
    num_included_in_starting_items: int = 0
    included_ammo: Tuple[int, ...] = tuple()
    allowed_as_random_starting_item: bool = True

    @property
    def as_json(self) -> dict:
        result = {}

        for field in dataclasses.fields(self):
            value = getattr(self, field.name)
            if value != field.default:
                result[field.name] = value

        if "included_ammo" in result:
            result["included_ammo"] = list(result["included_ammo"])

        return result

    @classmethod
    def from_json(cls, value: dict) -> "MajorItemState":
        kwargs = {}

        for field in dataclasses.fields(cls):
            if field.name in value:
                kwargs[field.name] = value[field.name]

        if "included_ammo" in kwargs:
            kwargs["included_ammo"] = tuple(kwargs["included_ammo"])

        return cls(**kwargs)

    def bit_pack_encode(self, item: MajorItem) -> Iterator[Tuple[int, int]]:
        # original location
        yield int(self.include_copy_in_original_location), 2

        # num shuffled
        yield from bitpacking.encode_int_with_limits(self.num_shuffled_pickups, DEFAULT_MAXIMUM_SHUFFLED)

        # starting item
        yield self.num_included_in_starting_items, (
            ENERGY_TANK_MAXIMUM_COUNT if item.item_category == ItemCategory.ENERGY_TANK else 2)

        # ammo index
        assert len(self.included_ammo) == len(item.ammo_index)
        for ammo in self.included_ammo:
            yield ammo, 256

        # allowed_as_random_starting_item
        yield from bitpacking.encode_bool(self.allowed_as_random_starting_item)

    @classmethod
    def bit_pack_unpack(cls, decoder: BitPackDecoder, item: MajorItem) -> "MajorItemState":
        original = decoder.decode_single(2)

        # num shuffled
        shuffled = bitpacking.decode_int_with_limits(decoder, DEFAULT_MAXIMUM_SHUFFLED)

        # starting item
        starting = decoder.decode_single(ENERGY_TANK_MAXIMUM_COUNT if item.item_category == ItemCategory.ENERGY_TANK
                                         else 2)

        if item.ammo_index:
            included_ammo = decoder.decode(*[256 for _ in item.ammo_index])
        else:
            included_ammo = []

        # allowed_as_random_starting_item
        allowed_as_random_starting_item = bitpacking.decode_bool(decoder)

        return cls(
            include_copy_in_original_location=bool(original),
            num_shuffled_pickups=shuffled,
            num_included_in_starting_items=starting,
            included_ammo=tuple(included_ammo),
            allowed_as_random_starting_item=allowed_as_random_starting_item,
        )



from typing import List

from randovania.game_description.resources.pickup_entry import PickupEntry
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.generator.item_pool import PoolResults, pickup_creator


def add_energy_cells(resource_database: ResourceDatabase,
                     ) -> PoolResults:
    """
    :param resource_database:
    :return:
    """
    item_pool: List[PickupEntry] = []

    for i in range(9):
        item_pool.append(pickup_creator.create_energy_cell(i, resource_database))

    return PoolResults(item_pool, {}, {})



from unittest.mock import MagicMock, call

import pytest
from mock import AsyncMock

from randovania.game_connection.connection_backend import ConnectionBackend
from randovania.game_connection.dolphin_backend import DolphinBackend
from randovania.game_connection.game_connection import GameConnection


@pytest.fixture(name="connection")
def _connection(skip_qtbot):
    return GameConnection(MagicMock())


@pytest.mark.asyncio
async def test_set_backend(skip_qtbot):
    # Setup
    GameConnection._notify_status = MagicMock()
    backend1 = ConnectionBackend()
    backend2 = ConnectionBackend()
    game_connection = GameConnection(backend1)

    listener = AsyncMock()

    # Run
    game_connection.set_location_collected_listener(listener)
    await backend1._emit_location_collected(5)
    await backend2._emit_location_collected(6)
    game_connection.set_backend(backend2)
    await backend1._emit_location_collected(7)
    await backend2._emit_location_collected(8)

    # Assert
    listener.assert_has_awaits([call(5), call(8)])
    game_connection._notify_status.assert_has_calls([call(), call()])


@pytest.mark.asyncio
async def test_update(skip_qtbot, qapp):
    # Setup
    backend_update = AsyncMock()

    backend = MagicMock()
    backend.update = backend_update

    game_connection = GameConnection(backend)
    game_connection._notify_status = MagicMock()

    # Run
    await game_connection._update()

    # Assert
    backend_update.assert_awaited_once_with(game_connection._dt)
    game_connection._notify_status.assert_called_once_with()


def test_pretty_current_status(skip_qtbot):
    # Setup
    connection = GameConnection(DolphinBackend())

    # Run
    result = connection.pretty_current_status

    # Assert
    assert result == f"Dolphin: Disconnected"


def test_get_current_inventory(connection):
    result = connection.get_current_inventory()

    assert result is connection.backend.get_current_inventory.return_value



import copy
import dataclasses
from random import Random

from randovania.game_description import data_reader
from randovania.game_description.area_location import AreaLocation
from randovania.game_description.assignment import GateAssignment
from randovania.game_description.echoes_game_specific import EchoesGameSpecific
from randovania.game_description.game_description import GameDescription
from randovania.game_description.game_patches import GamePatches
from randovania.game_description.hint import Hint, HintType, PrecisionPair, HintLocationPrecision, HintItemPrecision, \
    HintDarkTemple
from randovania.game_description.node import LogbookNode, LoreType
from randovania.game_description.resources.logbook_asset import LogbookAsset
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.resources.resource_database import ResourceDatabase
from randovania.game_description.resources.resource_type import ResourceType
from randovania.game_description.world_list import WorldList
from randovania.games.game import RandovaniaGame
from randovania.generator import elevator_distributor
from randovania.interface_common.enum_lib import iterate_enum
from randovania.layout.echoes_configuration import EchoesConfiguration
from randovania.layout.elevators import LayoutElevators
from randovania.layout.translator_configuration import LayoutTranslatorRequirement


class MissingRng(Exception):
    pass


# M-Dhe -> E3B417BF -> Temple Grounds - Landing Site -> Defiled Shrine -> 11
# J-Fme -> 65206511 -> Temple Grounds - Industrial Site -> Accursed Lake -> 15
# D-Isl -> 28E8C41A -> Temple Grounds - Storage Cavern A -> Ing Reliquary -> 19
# J-Stl -> 150E8DB8 -> Agon Wastes - Central Mining Station -> Battleground -> 45
# B-Stl -> DE525E1D -> Agon Wastes - Main Reactor -> Dark Oasis -> 53
# S-Dly -> 58C62CB3 -> Torvus Bog - Torvus Lagoon -> Poisoned Bog -> 68
# G-Sch -> 939AFF16 -> Torvus Bog - Catacombs -> Dungeon -> 91
# C-Rch -> A9909E66 -> Sanctuary Fortress - Dynamo Works -> Hive Dynamo Works -> 106
# S-Jrs -> 62CC4DC3 -> Sanctuary Fortress - Sanctuary Entrance -> Hive Entrance -> 117

_KEYBEARERS_HINTS = {
    LogbookAsset(0xDE525E1D): PickupIndex(53),
    LogbookAsset(0xA9909E66): PickupIndex(106),
    LogbookAsset(0x28E8C41A): PickupIndex(19),
    LogbookAsset(0x939AFF16): PickupIndex(91),
    LogbookAsset(0x65206511): PickupIndex(15),
    LogbookAsset(0x150E8DB8): PickupIndex(45),
    LogbookAsset(0xE3B417BF): PickupIndex(11),
    LogbookAsset(0x58C62CB3): PickupIndex(68),
    LogbookAsset(0x62CC4DC3): PickupIndex(117),
}


def add_elevator_connections_to_patches(layout_configuration: EchoesConfiguration,
                                        rng: Random,
                                        patches: GamePatches) -> GamePatches:
    """
    :param layout_configuration:
    :param rng:
    :param patches:
    :return:
    """
    elevator_connection = copy.copy(patches.elevator_connection)

    if layout_configuration.elevators != LayoutElevators.VANILLA:
        if rng is None:
            raise MissingRng("Elevator")

        world_list = data_reader.decode_data(layout_configuration.game_data).world_list
        areas_to_not_change = {
            2278776548,  # Sky Temple Gateway
            2068511343,  # Sky Temple Energy Controller
            3136899603,  # Aerie Transport Station
            1564082177,  # Aerie
        }

        elevator_db = elevator_distributor.create_elevator_database(world_list, areas_to_not_change)

        if layout_configuration.elevators in {LayoutElevators.TWO_WAY_RANDOMIZED, LayoutElevators.TWO_WAY_UNCHECKED}:
            connections = elevator_distributor.two_way_elevator_connections(
                rng=rng,
                elevator_database=elevator_db,
                between_areas=layout_configuration.elevators == LayoutElevators.TWO_WAY_RANDOMIZED
            )
        else:
            connections = elevator_distributor.one_way_elevator_connections(
                rng=rng,
                elevator_database=elevator_db,
                world_list=world_list,
                elevator_target=layout_configuration.elevators == LayoutElevators.ONE_WAY_ELEVATOR
            )

        elevator_connection.update(connections)

    if layout_configuration.skip_final_bosses:
        elevator_connection[136970379] = AreaLocation(1006255871, 1393588666)

    return dataclasses.replace(patches, elevator_connection=elevator_connection)


def gate_assignment_for_configuration(configuration: EchoesConfiguration,
                                      resource_database: ResourceDatabase,
                                      rng: Random,
                                      ) -> GateAssignment:
    """
    :param configuration:
    :param resource_database:
    :param rng:
    :return:
    """

    all_choices = list(LayoutTranslatorRequirement)
    all_choices.remove(LayoutTranslatorRequirement.RANDOM)
    all_choices.remove(LayoutTranslatorRequirement.RANDOM_WITH_REMOVED)
    without_removed = copy.copy(all_choices)
    without_removed.remove(LayoutTranslatorRequirement.REMOVED)
    random_requirements = {LayoutTranslatorRequirement.RANDOM, LayoutTranslatorRequirement.RANDOM_WITH_REMOVED}

    result = {}
    for gate, requirement in configuration.translator_configuration.translator_requirement.items():
        if requirement in random_requirements:
            if rng is None:
                raise MissingRng("Translator")
            requirement = rng.choice(all_choices if requirement == LayoutTranslatorRequirement.RANDOM_WITH_REMOVED
                                     else without_removed)

        result[gate] = resource_database.get_by_type_and_index(ResourceType.ITEM, requirement.item_index)

    return result


def starting_location_for_configuration(configuration: EchoesConfiguration,
                                        game: GameDescription,
                                        rng: Random,
                                        ) -> AreaLocation:
    locations = list(configuration.starting_location.locations)
    if len(locations) == 0:
        raise ValueError("No available starting locations")
    elif len(locations) == 1:
        location = locations[0]
    else:
        if rng is None:
            raise MissingRng("Starting Location")
        location = rng.choice(locations)

    return location


def add_echoes_default_hints_to_patches(rng: Random,
                                        patches: GamePatches,
                                        world_list: WorldList,
                                        num_joke: int,
                                        is_multiworld: bool,
                                        ) -> GamePatches:
    """
    Adds hints that are present on all games.
    :param rng:
    :param patches:
    :param world_list:
    :param num_joke
    :param is_multiworld
    :return:
    """

    for node in world_list.all_nodes:
        if isinstance(node, LogbookNode) and node.lore_type == LoreType.LUMINOTH_WARRIOR:
            patches = patches.assign_hint(node.resource(),
                                          Hint(HintType.LOCATION,
                                               PrecisionPair(HintLocationPrecision.KEYBEARER,
                                                             HintItemPrecision.BROAD_CATEGORY,
                                                             include_owner=True),
                                               PickupIndex(node.hint_index)))

    all_logbook_assets = [node.resource()
                          for node in world_list.all_nodes
                          if isinstance(node, LogbookNode)
                          and node.resource() not in patches.hints
                          and node.lore_type.holds_generic_hint]

    rng.shuffle(all_logbook_assets)

    # The 4 guaranteed hints
    indices_with_hint = [
        (PickupIndex(24), HintLocationPrecision.LIGHT_SUIT_LOCATION),  # Light Suit
        (PickupIndex(43), HintLocationPrecision.GUARDIAN),  # Dark Suit (Amorbis)
        (PickupIndex(79), HintLocationPrecision.GUARDIAN),  # Dark Visor (Chykka)
        (PickupIndex(115), HintLocationPrecision.GUARDIAN),  # Annihilator Beam (Quadraxis)
    ]
    rng.shuffle(indices_with_hint)
    for index, location_type in indices_with_hint:
        if not all_logbook_assets:
            break

        logbook_asset = all_logbook_assets.pop()
        patches = patches.assign_hint(logbook_asset, Hint(HintType.LOCATION,
                                                          PrecisionPair(location_type, HintItemPrecision.DETAILED,
                                                                        include_owner=False),
                                                          index))

    # Dark Temple hints
    temple_hints = list(iterate_enum(HintDarkTemple))
    while all_logbook_assets and temple_hints:
        logbook_asset = all_logbook_assets.pop()
        patches = patches.assign_hint(logbook_asset, Hint(HintType.RED_TEMPLE_KEY_SET, None,
                                                          dark_temple=temple_hints.pop(0)))

    # Jokes
    while num_joke > 0 and all_logbook_assets:
        logbook_asset = all_logbook_assets.pop()
        patches = patches.assign_hint(logbook_asset, Hint(HintType.JOKE, None))
        num_joke -= 1

    return patches


def create_game_specific(configuration: EchoesConfiguration, game: GameDescription) -> EchoesGameSpecific:
    if configuration.game == RandovaniaGame.PRIME2:
        return EchoesGameSpecific(
            energy_per_tank=configuration.energy_per_tank,
            safe_zone_heal_per_second=configuration.safe_zone.heal_per_second,
            beam_configurations=configuration.beam_configuration.create_game_specific(game.resource_database),
            dangerous_energy_tank=configuration.dangerous_energy_tank,
        )
    else:
        return game.game_specific


def create_base_patches(configuration: EchoesConfiguration,
                        rng: Random,
                        game: GameDescription,
                        is_multiworld: bool,
                        ) -> GamePatches:
    """
    """
    patches = dataclasses.replace(game.create_game_patches(),
                                  game_specific=create_game_specific(configuration, game))

    patches = add_elevator_connections_to_patches(configuration, rng, patches)

    # Gates
    if configuration.game == RandovaniaGame.PRIME2:
        patches = patches.assign_gate_assignment(
            gate_assignment_for_configuration(configuration, game.resource_database, rng))

    # Starting Location
    patches = patches.assign_starting_location(
        starting_location_for_configuration(configuration, game, rng))

    # Hints
    if rng is not None and configuration.game == RandovaniaGame.PRIME2:
        patches = add_echoes_default_hints_to_patches(rng, patches, game.world_list,
                                                      num_joke=2, is_multiworld=is_multiworld)

    return patches



import copy
import dataclasses
import json
from random import Random
from typing import Tuple
from unittest.mock import MagicMock, patch

import pytest

import randovania
from randovania.game_description import data_reader
from randovania.game_description.area_location import AreaLocation
from randovania.game_description.assignment import PickupTarget
from randovania.game_description.default_database import default_prime2_memo_data
from randovania.game_description.item.item_category import ItemCategory
from randovania.game_description.resources.item_resource_info import ItemResourceInfo
from randovania.game_description.resources.pickup_entry import ConditionalResources, PickupEntry, ResourceConversion
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.resources.resource_type import ResourceType
from randovania.game_description.resources.simple_resource_info import SimpleResourceInfo
from randovania.game_description.resources.translator_gate import TranslatorGate
from randovania.games.prime import patcher_file, default_data
from randovania.generator.item_pool import pickup_creator, pool_creator
from randovania.interface_common.cosmetic_patches import CosmeticPatches
from randovania.interface_common.players_configuration import PlayersConfiguration
from randovania.layout.hint_configuration import SkyTempleKeyHintMode, HintConfiguration
from randovania.layout.elevators import LayoutElevators
from randovania.layout.layout_description import LayoutDescription
from randovania.layout.major_item_state import MajorItemState
from randovania.layout.pickup_model import PickupModelStyle, PickupModelDataSource


def test_add_header_data_to_result():
    # Setup
    description = MagicMock()
    description.shareable_word_hash = "<shareable_word_hash>"
    description.shareable_hash = "<shareable_hash>"
    expected = {
        "permalink": "-permalink-",
        "seed_hash": "- <shareable_word_hash> (<shareable_hash>)",
        "randovania_version": randovania.VERSION,
    }
    result = {}

    # Run
    patcher_file._add_header_data_to_result(description, result)

    # Assert
    assert json.loads(json.dumps(result)) == expected


def test_create_spawn_point_field(echoes_resource_database, empty_patches):
    # Setup
    patches = empty_patches.assign_starting_location(AreaLocation(100, 5000)).assign_extra_initial_items({
        echoes_resource_database.get_by_type_and_index(ResourceType.ITEM, 15): 3,
    })

    capacities = [
        {'amount': 3 if item.index == 15 else 0, 'index': item.index}
        for item in echoes_resource_database.item
    ]

    # Run
    result = patcher_file._create_spawn_point_field(patches, echoes_resource_database)

    # Assert
    assert result == {
        "location": {
            "world_asset_id": 100,
            "area_asset_id": 5000,
        },
        "amount": capacities,
        "capacity": capacities,
    }


def test_create_elevators_field_no_elevator(empty_patches):
    # Setup
    game = data_reader.decode_data(default_data.decode_default_prime2())

    # Run
    with pytest.raises(ValueError) as exp:
        patcher_file._create_elevators_field(empty_patches, game)

    # Assert
    assert str(exp.value) == "Invalid elevator count. Expected 22, got 0."


@pytest.mark.parametrize("vanilla_gateway", [False, True])
def test_create_elevators_field_elevators_for_a_seed(vanilla_gateway: bool,
                                                     echoes_resource_database, empty_patches):
    # Setup
    game = data_reader.decode_data(default_data.decode_default_prime2())
    patches = game.create_game_patches()

    elevator_connection = copy.copy(patches.elevator_connection)
    elevator_connection[589851] = AreaLocation(464164546, 900285955)
    elevator_connection[1572998] = AreaLocation(1039999561, 3479543630)

    if not vanilla_gateway:
        elevator_connection[136970379] = AreaLocation(2252328306, 3619928121)

    patches = dataclasses.replace(patches, elevator_connection=elevator_connection)

    # Run
    result = patcher_file._create_elevators_field(patches, game)

    # Assert
    expected = [
        {"instance_id": 589851,
         "origin_location": {"world_asset_id": 1006255871, "area_asset_id": 2918020398},
         "target_location": {"world_asset_id": 464164546, "area_asset_id": 900285955},
         "room_name": "Transport to Sanctuary Spider side"},

        {"instance_id": 1572998,
         "origin_location": {"world_asset_id": 1006255871, "area_asset_id": 1660916974},
         "target_location": {"world_asset_id": 1039999561, "area_asset_id": 3479543630},
         "room_name": "Transport to Torvus Temple Access"},

        {"instance_id": 1966093,
         "origin_location": {"world_asset_id": 1006255871, "area_asset_id": 2889020216},
         "target_location": {"world_asset_id": 1039999561, "area_asset_id": 1868895730},
         "room_name": "Transport to Torvus Entrance"},

        {"instance_id": 2097251,
         "origin_location": {"world_asset_id": 1006255871, "area_asset_id": 1287880522},
         "target_location": {"world_asset_id": 2252328306, "area_asset_id": 2399252740},
         "room_name": "Transport to Temple Transport Violet"},

        {"instance_id": 136970379,
         "origin_location": {"world_asset_id": 1006255871, "area_asset_id": 2278776548},
         "target_location": {"world_asset_id": 2252328306,
                             "area_asset_id": 2068511343 if vanilla_gateway else 3619928121},
         "room_name": "Sky Temple Gateway" if vanilla_gateway else "Transport to Sanctum"},

        {"instance_id": 3342446,
         "origin_location": {"world_asset_id": 1006255871, "area_asset_id": 3455543403},
         "target_location": {"world_asset_id": 464164546, "area_asset_id": 3528156989},
         "room_name": "Transport to Sanctuary Entrance"},

        {"instance_id": 3538975,
         "origin_location": {"world_asset_id": 1006255871, "area_asset_id": 1345979968},
         "target_location": {"world_asset_id": 2252328306, "area_asset_id": 408633584},
         "room_name": "Transport to Temple Transport Emerald"},

        {"instance_id": 152,
         "origin_location": {"world_asset_id": 2252328306, "area_asset_id": 408633584},
         "target_location": {"world_asset_id": 1006255871, "area_asset_id": 1345979968},
         "room_name": "Transport to Sanctuary Quadrant"},

        {"instance_id": 393260,
         "origin_location": {"world_asset_id": 2252328306, "area_asset_id": 2556480432},
         "target_location": {"world_asset_id": 1006255871, "area_asset_id": 2918020398},
         "room_name": "Transport to Torvus Quadrant"},

        {"instance_id": 524321,
         "origin_location": {"world_asset_id": 2252328306, "area_asset_id": 2399252740},
         "target_location": {"world_asset_id": 1006255871, "area_asset_id": 1287880522},
         "room_name": "Transport to Agon Quadrant"},

        {"instance_id": 589949,
         "origin_location": {"world_asset_id": 2252328306, "area_asset_id": 2068511343},
         "target_location": {"world_asset_id": 1006255871, "area_asset_id": 2278776548},
         "room_name": "Sky Temple Energy Controller"},

        {"instance_id": 122,
         "origin_location": {"world_asset_id": 1119434212, "area_asset_id": 1473133138},
         "target_location": {"world_asset_id": 1006255871, "area_asset_id": 1660916974},
         "room_name": "Transport to Agon Gate"},

        {"instance_id": 1245307,
         "origin_location": {"world_asset_id": 1119434212, "area_asset_id": 2806956034},
         "target_location": {"world_asset_id": 1039999561, "area_asset_id": 3479543630},
         "room_name": "Transport to Torvus Temple Access"},

        {"instance_id": 2949235,
         "origin_location": {"world_asset_id": 1119434212, "area_asset_id": 3331021649},
         "target_location": {"world_asset_id": 464164546, "area_asset_id": 900285955},
         "room_name": "Transport to Sanctuary Spider side"},

        {"instance_id": 129,
         "origin_location": {"world_asset_id": 1039999561, "area_asset_id": 1868895730},
         "target_location": {"world_asset_id": 1006255871, "area_asset_id": 2889020216},
         "room_name": "Transport to Torvus Gate"},

        {"instance_id": 2162826,
         "origin_location": {"world_asset_id": 1039999561, "area_asset_id": 3479543630},
         "target_location": {"world_asset_id": 1119434212, "area_asset_id": 2806956034},
         "room_name": "Transport to Agon Portal Access"},

        {"instance_id": 4522032,
         "origin_location": {"world_asset_id": 1039999561, "area_asset_id": 3205424168},
         "target_location": {"world_asset_id": 464164546, "area_asset_id": 3145160350},
         "room_name": "Transport to Sanctuary Vault side"},

        {"instance_id": 38,
         "origin_location": {"world_asset_id": 464164546, "area_asset_id": 3528156989},
         "target_location": {"world_asset_id": 1006255871, "area_asset_id": 3455543403},
         "room_name": "Transport to Sanctuary Gate"},

        {"instance_id": 1245332,
         "origin_location": {"world_asset_id": 464164546, "area_asset_id": 900285955},
         "target_location": {"world_asset_id": 1119434212, "area_asset_id": 3331021649},
         "room_name": "Transport to Agon Temple Access"},

        {"instance_id": 1638535,
         "origin_location": {"world_asset_id": 464164546, "area_asset_id": 3145160350},
         "target_location": {"world_asset_id": 1039999561, "area_asset_id": 3205424168},
         "room_name": "Transport to Lower Torvus Access"},

        {'instance_id': 204865660,
         'origin_location': {'world_asset_id': 464164546, 'area_asset_id': 3136899603},
         'target_location': {'world_asset_id': 464164546, 'area_asset_id': 1564082177},
         'room_name': 'Aerie Transport Station'},

        {'instance_id': 4260106,
         'origin_location': {'world_asset_id': 464164546, 'area_asset_id': 1564082177},
         'target_location': {'world_asset_id': 464164546, 'area_asset_id': 3136899603},
         'room_name': 'Aerie'}
    ]
    assert result == expected


def test_create_translator_gates_field():
    # Setup
    gate_assignment = {
        TranslatorGate(1): SimpleResourceInfo(10, "LongA", "A", ResourceType.ITEM),
        TranslatorGate(3): SimpleResourceInfo(50, "LongB", "B", ResourceType.ITEM),
        TranslatorGate(4): SimpleResourceInfo(10, "LongA", "A", ResourceType.ITEM),
    }

    # Run
    result = patcher_file._create_translator_gates_field(gate_assignment)

    # Assert
    assert result == [
        {"gate_index": 1, "translator_index": 10},
        {"gate_index": 3, "translator_index": 50},
        {"gate_index": 4, "translator_index": 10},
    ]


@pytest.mark.parametrize("elevators", [LayoutElevators.VANILLA, LayoutElevators.TWO_WAY_RANDOMIZED])
def test_apply_translator_gate_patches(elevators):
    # Setup
    target = {}

    # Run
    patcher_file._apply_translator_gate_patches(target, elevators)

    # Assert
    assert target == {
        "always_up_gfmc_compound": True,
        "always_up_torvus_temple": True,
        "always_up_great_temple": elevators != LayoutElevators.VANILLA,
    }


def test_get_single_hud_text_locked_pbs():
    # Run
    result = patcher_file._get_single_hud_text("Temporary Power Bombs", patcher_file._simplified_memo_data(), tuple())

    # Assert
    assert result == "Power Bomb Expansion acquired, but the main Power Bomb is required to use it."


@pytest.mark.parametrize("order", [
    ("X", "Y"),
    ("Y", "X"),
    ("Y", "Z"),
])
def test_calculate_hud_text(order: Tuple[str, str]):
    # Setup
    resource_a = ItemResourceInfo(1, "A", "A", 10, None)
    resource_b = ItemResourceInfo(2, "B", "B", 10, None)

    pickup_x = PickupEntry("A", 1, ItemCategory.TEMPLE_KEY, ItemCategory.KEY,
                           (
                               ConditionalResources("A", None, ((resource_a, 1),)),
                           ))
    pickup_y = PickupEntry("Y", 2, ItemCategory.SUIT, ItemCategory.LIFE_SUPPORT,
                           (
                               ConditionalResources("B", None, ((resource_b, 1),)),
                               ConditionalResources("A", resource_b, ((resource_a, 5),))
                           ))
    pickup_z = PickupEntry("Z", 2, ItemCategory.SUIT, ItemCategory.LIFE_SUPPORT,
                           (
                               ConditionalResources("A", None, ((resource_a, 1),)),
                               ConditionalResources("B", resource_a, ((resource_b, 5),))
                           ))

    memo_data = {
        "A": "You got {A} of A",
        "B": "You found {B} of B",
    }
    pickups = {
        "X": pickup_x,
        "Y": pickup_y,
        "Z": pickup_z,
    }

    # Run
    result = patcher_file._calculate_hud_text(pickups[order[0]], pickups[order[1]], PickupModelStyle.HIDE_ALL,
                                              memo_data)

    # Assert
    if order[1] == "Y":
        assert result == ["You found 1 of B"]
    elif order[1] == "X":
        assert result == ["You got 1 of A", "You got 1 of A"]
    else:
        assert result == ["You got 1 of A", "You found 5 of B"]


@pytest.mark.parametrize("model_style", PickupModelStyle)
def test_create_pickup_list(model_style: PickupModelStyle, empty_patches):
    # Setup
    has_scan_text = model_style in {PickupModelStyle.ALL_VISIBLE, PickupModelStyle.HIDE_MODEL}
    rng = Random(5000)

    useless_resource = ItemResourceInfo(0, "Useless", "Useless", 10, None)
    resource_a = ItemResourceInfo(1, "A", "A", 10, None)
    resource_b = ItemResourceInfo(2, "B", "B", 10, None)
    pickup_a = PickupEntry("A", 1, ItemCategory.TEMPLE_KEY, ItemCategory.KEY,
                           (
                               ConditionalResources(None, None, ((resource_a, 1),)),
                           ))
    pickup_b = PickupEntry("B", 2, ItemCategory.SUIT, ItemCategory.LIFE_SUPPORT,
                           (
                               ConditionalResources(None, None, ((resource_b, 1), (resource_a, 1))),
                               ConditionalResources(None, resource_b, ((resource_a, 5),))
                           ))
    pickup_c = PickupEntry("C", 2, ItemCategory.EXPANSION, ItemCategory.MISSILE_RELATED,
                           resources=(
                               ConditionalResources(None, None, ((resource_b, 2), (resource_a, 1))),
                           ),
                           convert_resources=(
                               ResourceConversion(useless_resource, resource_a),
                           ))

    useless_pickup = PickupEntry("Useless", 0, ItemCategory.ETM, ItemCategory.ETM,
                                 (
                                     ConditionalResources(None, None, ((useless_resource, 1),)),
                                 ))
    patches = empty_patches.assign_pickup_assignment({
        PickupIndex(0): PickupTarget(pickup_a, 0),
        PickupIndex(2): PickupTarget(pickup_b, 0),
        PickupIndex(3): PickupTarget(pickup_a, 0),
        PickupIndex(4): PickupTarget(pickup_c, 0),
    })
    creator = patcher_file.PickupCreatorSolo(MagicMock(), patcher_file._SimplifiedMemo())

    # Run
    result = patcher_file._create_pickup_list(patches,
                                              PickupTarget(useless_pickup, 0),
                                              5,
                                              rng,
                                              model_style,
                                              PickupModelDataSource.ETM,
                                              creator,
                                              )

    # Assert
    assert len(result) == 5
    assert result[0] == {
        "pickup_index": 0,
        "model_index": 1 if model_style == PickupModelStyle.ALL_VISIBLE else 30,
        "scan": "A" if has_scan_text else "Unknown item",
        "hud_text": ["A acquired!"] if model_style != PickupModelStyle.HIDE_ALL else ['Unknown item acquired!'],
        "sound_index": 1 if model_style == PickupModelStyle.ALL_VISIBLE else 0,
        "jingle_index": 2 if model_style == PickupModelStyle.ALL_VISIBLE else 0,
        "resources": [
            {
                "index": 1,
                "amount": 1
            }
        ],
        "conditional_resources": [],
        "convert": [],
    }
    assert result[1] == {
        "pickup_index": 1,
        "scan": "Useless" if has_scan_text else "Unknown item",
        "model_index": 0 if model_style == PickupModelStyle.ALL_VISIBLE else 30,
        "hud_text": ["Useless acquired!"] if model_style != PickupModelStyle.HIDE_ALL else ['Unknown item acquired!'],
        "sound_index": 0,
        "jingle_index": 0,
        "resources": [
            {
                "index": 0,
                "amount": 1
            }
        ],
        "conditional_resources": [],
        "convert": [],
    }
    assert result[2] == {
        "pickup_index": 2,
        "scan": "B" if has_scan_text else "Unknown item",
        "model_index": 2 if model_style == PickupModelStyle.ALL_VISIBLE else 30,
        "hud_text": ["B acquired!", "B acquired!"] if model_style != PickupModelStyle.HIDE_ALL else [
            'Unknown item acquired!', 'Unknown item acquired!'],
        "sound_index": 0,
        "jingle_index": 1 if model_style == PickupModelStyle.ALL_VISIBLE else 0,
        "resources": [
            {
                "index": 2,
                "amount": 1
            },
            {
                "index": 1,
                "amount": 1
            }
        ],
        "conditional_resources": [{
            "item": 2,
            "resources": [
                {
                    "index": 1,
                    "amount": 5
                }
            ]
        }],
        "convert": [],
    }
    assert result[3] == {
        "pickup_index": 3,
        "scan": "A" if has_scan_text else "Unknown item",
        "model_index": 1 if model_style == PickupModelStyle.ALL_VISIBLE else 30,
        "hud_text": ["A acquired!"] if model_style != PickupModelStyle.HIDE_ALL else ['Unknown item acquired!'],
        "sound_index": 1 if model_style == PickupModelStyle.ALL_VISIBLE else 0,
        "jingle_index": 2 if model_style == PickupModelStyle.ALL_VISIBLE else 0,
        "resources": [
            {
                "index": 1,
                "amount": 1
            }
        ],
        "conditional_resources": [],
        "convert": [],
    }
    assert result[4] == {
        "pickup_index": 4,
        "scan": "C that provides 2 B and 1 A" if has_scan_text else "Unknown item",
        "model_index": 2 if model_style == PickupModelStyle.ALL_VISIBLE else 30,
        "hud_text": ["C acquired!"] if model_style != PickupModelStyle.HIDE_ALL else ['Unknown item acquired!'],
        "sound_index": 0,
        "jingle_index": 0,
        "resources": [
            {
                "index": 2,
                "amount": 2
            },
            {
                "index": 1,
                "amount": 1
            }
        ],
        "conditional_resources": [],
        "convert": [
            {
                "from_item": 0,
                "to_item": 1,
                "clear_source": True,
                "overwrite_target": False,
            }
        ],
    }


@pytest.mark.parametrize("has_memo_data", [False, True])
def test_create_pickup_list_random_data_source(has_memo_data: bool, empty_patches):
    # Setup
    rng = Random(5000)
    resources = (ConditionalResources(None, None, tuple()),)
    resource_b = ItemResourceInfo(2, "B", "B", 10, None)

    pickup_a = PickupEntry("A", 1, ItemCategory.TEMPLE_KEY, ItemCategory.KEY, resources)
    pickup_b = PickupEntry("B", 2, ItemCategory.SUIT, ItemCategory.LIFE_SUPPORT,
                           (ConditionalResources(None, None, tuple()),
                            ConditionalResources(None, resource_b, tuple()),))
    pickup_c = PickupEntry("C", 2, ItemCategory.EXPANSION, ItemCategory.MISSILE_RELATED, resources)
    useless_pickup = PickupEntry("Useless", 0, ItemCategory.ETM, ItemCategory.ETM, resources)

    patches = empty_patches.assign_pickup_assignment({
        PickupIndex(0): PickupTarget(pickup_a, 0),
        PickupIndex(2): PickupTarget(pickup_b, 0),
        PickupIndex(3): PickupTarget(pickup_a, 0),
        PickupIndex(4): PickupTarget(pickup_c, 0),
    })

    if has_memo_data:
        memo_data = {
            "A": "This is an awesome item A",
            "B": "This is B. It is good.",
            "C": "What a nice day to have a C",
            "Useless": "Try again next time",
        }
    else:
        memo_data = {
            name: "{} acquired!".format(name)
            for name in ("A", "B", "C", "Useless")
        }

    creator = patcher_file.PickupCreatorSolo(MagicMock(), memo_data)

    # Run
    result = patcher_file._create_pickup_list(patches,
                                              PickupTarget(useless_pickup, 0),
                                              5,
                                              rng,
                                              PickupModelStyle.HIDE_ALL,
                                              PickupModelDataSource.RANDOM,
                                              creator,
                                              )

    # Assert
    # Assert
    assert len(result) == 5
    assert result[0] == {
        "pickup_index": 0,
        "model_index": 1,
        "scan": "A",
        "hud_text": [memo_data["A"]],
        "sound_index": 1,
        "jingle_index": 2,
        "resources": [],
        "conditional_resources": [],
        "convert": [],
    }
    assert result[1] == {
        "pickup_index": 1,
        "scan": "A",
        "model_index": 1,
        "hud_text": [memo_data["A"]],
        "sound_index": 1,
        "jingle_index": 2,
        "resources": [],
        "conditional_resources": [],
        "convert": [],
    }
    assert result[2] == {
        "pickup_index": 2,
        "scan": "C",
        "model_index": 2,
        "hud_text": [memo_data["C"], memo_data["C"]],
        "sound_index": 0,
        "jingle_index": 0,
        "resources": [],
        "conditional_resources": [{'item': 2, 'resources': []}],
        "convert": [],
    }
    assert result[3] == {
        "pickup_index": 3,
        "scan": "B",
        "model_index": 2,
        "hud_text": [memo_data["B"]],
        "sound_index": 0,
        "jingle_index": 1,
        "resources": [],
        "conditional_resources": [],
        "convert": [],
    }
    assert result[4] == {
        "pickup_index": 4,
        "scan": "A",
        "model_index": 1,
        "hud_text": [memo_data["A"]],
        "sound_index": 1,
        "jingle_index": 2,
        "resources": [],
        "conditional_resources": [],
        "convert": [],
    }


def test_pickup_scan_for_progressive_suit(echoes_item_database, echoes_resource_database):
    # Setup
    progressive_suit = echoes_item_database.major_items["Progressive Suit"]
    pickup = pickup_creator.create_major_item(progressive_suit, MajorItemState(), False, echoes_resource_database,
                                              None, False)

    # Run
    result = patcher_file._pickup_scan(pickup)

    # Assert
    assert result == "Progressive Suit. Provides the following in order: Dark Suit, Light Suit"


@pytest.mark.parametrize("disable_hud_popup", [False, True])
def test_create_pickup_all_from_pool(echoes_resource_database,
                                     default_layout_configuration,
                                     disable_hud_popup: bool
                                     ):
    item_pool = pool_creator.calculate_pool_results(default_layout_configuration,
                                                    echoes_resource_database)
    index = PickupIndex(0)
    if disable_hud_popup:
        memo_data = patcher_file._SimplifiedMemo()
    else:
        memo_data = default_prime2_memo_data()
    creator = patcher_file.PickupCreatorSolo(MagicMock(), memo_data)

    for item in item_pool.pickups:
        creator.create_pickup(index, PickupTarget(item, 0), item, PickupModelStyle.ALL_VISIBLE)


def test_run_validated_hud_text():
    # Setup
    rng = MagicMock()
    rng.randint.return_value = 0
    creator = patcher_file.PickupCreatorSolo(rng, patcher_file._SimplifiedMemo())
    resource_a = SimpleResourceInfo(1, "A", "A", ResourceType.ITEM)
    pickup = PickupEntry("Energy Transfer Module", 1, ItemCategory.TEMPLE_KEY, ItemCategory.KEY,
                         (
                             ConditionalResources("Energy Transfer Module", None, ((resource_a, 1),)),
                         ))

    # Run
    data = creator.create_pickup_data(PickupIndex(0), PickupTarget(pickup, 0), pickup, PickupModelStyle.ALL_VISIBLE,
                                      "Scan Text")

    # Assert
    assert data['hud_text'] == ['Run validated!']


@pytest.fixture(name="pickup_for_create_pickup_data")
def _create_pickup_data():
    resource_a = ItemResourceInfo(1, "A", "A", 10, None)
    resource_b = ItemResourceInfo(2, "B", "B", 10, None)
    return PickupEntry("Cake", 1, ItemCategory.TEMPLE_KEY, ItemCategory.KEY,
                       (
                           ConditionalResources("Sugar", None, ((resource_a, 1),)),
                           ConditionalResources("Salt", resource_a, ((resource_b, 1),)),
                       ))


def test_solo_create_pickup_data(pickup_for_create_pickup_data):
    # Setup
    creator = patcher_file.PickupCreatorSolo(MagicMock(), patcher_file._SimplifiedMemo())

    # Run
    data = creator.create_pickup_data(PickupIndex(10), PickupTarget(pickup_for_create_pickup_data, 0),
                                      pickup_for_create_pickup_data, PickupModelStyle.ALL_VISIBLE,
                                      "Scan Text")

    # Assert
    assert data == {
        'conditional_resources': [
            {'item': 1, 'resources': [{'amount': 1, 'index': 2}]}
        ],
        'convert': [],
        'hud_text': ['Sugar acquired!', 'Salt acquired!'],
        'resources': [{'amount': 1, 'index': 1}],
        'scan': 'Scan Text',
    }


def test_multi_create_pickup_data_for_self(pickup_for_create_pickup_data):
    # Setup
    creator = patcher_file.PickupCreatorMulti(MagicMock(), patcher_file._SimplifiedMemo(),
                                              PlayersConfiguration(0, {0: "You",
                                                                       1: "Someone"}))

    # Run
    data = creator.create_pickup_data(PickupIndex(10), PickupTarget(pickup_for_create_pickup_data, 0),
                                      pickup_for_create_pickup_data, PickupModelStyle.ALL_VISIBLE,
                                      "Scan Text")

    # Assert
    assert data == {
        'conditional_resources': [
            {'item': 1, 'resources': [{'amount': 1, 'index': 2},
                                      {'amount': 11, 'index': 74},
                                      ]}
        ],
        'convert': [],
        'hud_text': ['Sugar acquired!', 'Salt acquired!'],
        'resources': [{'amount': 1, 'index': 1}, {'amount': 11, 'index': 74}, ],
        'scan': 'Your Scan Text',
    }


def test_multi_create_pickup_data_for_other(pickup_for_create_pickup_data):
    # Setup
    creator = patcher_file.PickupCreatorMulti(MagicMock(), patcher_file._SimplifiedMemo(),
                                              PlayersConfiguration(0, {0: "You",
                                                                       1: "Someone"}))

    # Run
    data = creator.create_pickup_data(PickupIndex(10), PickupTarget(pickup_for_create_pickup_data, 1),
                                      pickup_for_create_pickup_data, PickupModelStyle.ALL_VISIBLE,
                                      "Scan Text")

    # Assert
    assert data == {
        'conditional_resources': [],
        'convert': [],
        'hud_text': ['Sent Cake to Someone!'],
        'resources': [{'amount': 11, 'index': 74}, ],
        'scan': "Someone's Scan Text",
    }


@pytest.mark.parametrize("stk_mode", SkyTempleKeyHintMode)
@patch("randovania.games.prime.patcher_file._logbook_title_string_patches", autospec=True)
@patch("randovania.games.prime.patcher_file_lib.item_hints.create_hints", autospec=True)
@patch("randovania.games.prime.patcher_file_lib.sky_temple_key_hint.hide_hints", autospec=True)
@patch("randovania.games.prime.patcher_file_lib.sky_temple_key_hint.create_hints", autospec=True)
def test_create_string_patches(mock_stk_create_hints: MagicMock,
                               mock_stk_hide_hints: MagicMock,
                               mock_item_create_hints: MagicMock,
                               mock_logbook_title_string_patches: MagicMock,
                               stk_mode: SkyTempleKeyHintMode,
                               ):
    # Setup
    game = MagicMock()
    all_patches = MagicMock()
    rng = MagicMock()
    mock_item_create_hints.return_value = ["item", "hints"]
    mock_stk_create_hints.return_value = ["show", "hints"]
    mock_stk_hide_hints.return_value = ["hide", "hints"]
    player_config = PlayersConfiguration(0, {0: "you"})
    mock_logbook_title_string_patches.return_values = []

    # Run
    result = patcher_file._create_string_patches(HintConfiguration(sky_temple_keys=stk_mode),
                                                 game,
                                                 all_patches,
                                                 player_config,
                                                 rng,
                                                 )

    # Assert
    expected_result = ["item", "hints"]
    mock_item_create_hints.assert_called_once_with(all_patches, player_config, game.world_list, rng)
    mock_logbook_title_string_patches.assert_called_once_with()

    if stk_mode == SkyTempleKeyHintMode.DISABLED:
        mock_stk_hide_hints.assert_called_once_with()
        mock_stk_create_hints.assert_not_called()
        expected_result.extend(["hide", "hints"])

    else:
        mock_stk_create_hints.assert_called_once_with(all_patches, player_config, game.world_list,
                                                      stk_mode == SkyTempleKeyHintMode.HIDE_AREA)
        mock_stk_hide_hints.assert_not_called()
        expected_result.extend(["show", "hints"])

    assert result == expected_result


def test_create_patcher_file(test_files_dir):
    # Setup
    description = LayoutDescription.from_file(test_files_dir.joinpath("log_files", "seed_a.rdvgame"))
    player_index = 0
    preset = description.permalink.get_preset(player_index)
    cosmetic_patches = CosmeticPatches()

    # Run
    result = patcher_file.create_patcher_file(description, PlayersConfiguration(player_index, {0: "you"}),
                                              cosmetic_patches)

    # Assert
    assert isinstance(result["spawn_point"], dict)

    assert isinstance(result["pickups"], list)
    assert len(result["pickups"]) == 119

    assert isinstance(result["elevators"], list)
    assert len(result["elevators"]) == 22

    assert isinstance(result["translator_gates"], list)
    assert len(result["translator_gates"]) == 17

    assert isinstance(result["string_patches"], list)
    assert len(result["string_patches"]) == 60

    assert result["specific_patches"] == {
        "hive_chamber_b_post_state": True,
        "intro_in_post_state": True,
        "warp_to_start": preset.configuration.warp_to_start,
        "speed_up_credits": cosmetic_patches.speed_up_credits,
        "disable_hud_popup": cosmetic_patches.disable_hud_popup,
        "pickup_map_icons": cosmetic_patches.pickup_markers,
        "full_map_at_start": cosmetic_patches.open_map,
        "dark_world_varia_suit_damage": preset.configuration.varia_suit_damage,
        "dark_world_dark_suit_damage": preset.configuration.dark_suit_damage,
        "always_up_gfmc_compound": True,
        "always_up_torvus_temple": True,
        "always_up_great_temple": False,
    }
    assert result["default_items"] == {
        "visor": "Combat Visor",
        "beam": "Power Beam",
    }



import dataclasses
from random import Random
from typing import Tuple, List

import pytest

from randovania.game_description import data_reader
from randovania.game_description.area import Area
from randovania.game_description.dock import DockWeaknessDatabase
from randovania.game_description.echoes_game_specific import EchoesGameSpecific
from randovania.game_description.game_description import GameDescription
from randovania.game_description.node import ResourceNode, GenericNode, TranslatorGateNode
from randovania.game_description.requirements import Requirement
from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.game_description.resources.resource_info import add_resources_into_another
from randovania.game_description.resources.resource_type import ResourceType
from randovania.game_description.resources.translator_gate import TranslatorGate
from randovania.game_description.world import World
from randovania.game_description.world_list import WorldList
from randovania.games.game import RandovaniaGame
from randovania.generator import base_patches_factory, generator
from randovania.generator.generator_reach import GeneratorReach, filter_pickup_nodes, \
    reach_with_all_safe_resources, get_collectable_resource_nodes_of_reach, \
    advance_reach_with_possible_unsafe_resources, collectable_resource_nodes
from randovania.generator.item_pool import pool_creator
from randovania.layout.permalink import Permalink
from randovania.layout.preset import Preset
from randovania.layout.trick_level import LayoutTrickLevel, TrickLevelConfiguration
from randovania.resolver.bootstrap import logic_bootstrap
from randovania.resolver.state import State, add_pickup_to_state


def run_bootstrap(preset: Preset):
    game = data_reader.decode_data(preset.configuration.game_data)
    permalink = Permalink(
        seed_number=15000,
        spoiler=True,
        presets={0: preset},
    )
    patches = base_patches_factory.create_base_patches(preset.configuration, Random(15000), game, False)
    _, state = logic_bootstrap(preset.configuration, game, patches)

    return game, state, permalink


def _create_reach_with_unsafe(game: GameDescription, state: State) -> GeneratorReach:
    return advance_reach_with_possible_unsafe_resources(reach_with_all_safe_resources(game, state))


def _create_reaches_and_compare(game: GameDescription, state: State) -> Tuple[GeneratorReach, GeneratorReach]:
    first_reach = _create_reach_with_unsafe(game, state)
    second_reach = _create_reach_with_unsafe(game, first_reach.state)

    assert first_reach.is_safe_node(first_reach.state.node)
    assert second_reach.is_safe_node(first_reach.state.node)
    assert first_reach.is_safe_node(second_reach.state.node)
    assert second_reach.is_safe_node(second_reach.state.node)

    assert set(first_reach.safe_nodes) == set(second_reach.safe_nodes)
    assert set(first_reach.nodes) == set(second_reach.nodes)

    return first_reach, second_reach


def _compare_actions(first_reach: GeneratorReach,
                     second_reach: GeneratorReach,
                     ) -> Tuple[List[ResourceNode], List[ResourceNode]]:
    first_actions = get_collectable_resource_nodes_of_reach(first_reach)
    second_actions = get_collectable_resource_nodes_of_reach(second_reach)
    assert set(first_actions) == set(second_actions)

    return first_actions, second_actions


@pytest.mark.parametrize(("preset_name", "ignore_events", "ignore_pickups"), [
    ("Starter Preset", {91}, set()),  # Echoes
    ("Corruption Preset", {1, 146, 147, 148}, {0, 1, 2}),  # Corruption
    ("Prime Preset", {33}, set())  # Prime
])
def test_database_collectable(preset_manager, preset_name, ignore_events, ignore_pickups):
    game, initial_state, permalink = run_bootstrap(preset_manager.preset_for_name(preset_name).get_preset())
    all_pickups = set(filter_pickup_nodes(game.world_list.all_nodes))
    pool_results = pool_creator.calculate_pool_results(permalink.get_preset(0).configuration,
                                                       game.resource_database)
    add_resources_into_another(initial_state.resources, pool_results.initial_resources)
    for pickup in pool_results.pickups:
        add_pickup_to_state(initial_state, pickup)
    for pickup in pool_results.assignment.values():
        add_pickup_to_state(initial_state, pickup)
    for trick in game.resource_database.trick:
        initial_state.resources[trick] = LayoutTrickLevel.HYPERMODE.as_number

    expected_events = [event for event in game.resource_database.event if event.index not in ignore_events]
    expected_pickups = sorted(it.pickup_index for it in all_pickups if it.pickup_index.index not in ignore_pickups)

    reach = _create_reach_with_unsafe(game, initial_state.heal())
    while list(collectable_resource_nodes(reach.nodes, reach)):
        reach.act_on(next(iter(collectable_resource_nodes(reach.nodes, reach))))
        reach = advance_reach_with_possible_unsafe_resources(reach)

    # print("\nCurrent reach:")
    # for world in game.world_list.worlds:
    #     print(f"\n>> {world.name}")
    #     for node in world.all_nodes:
    #         print("[{!s:>5}, {!s:>5}, {!s:>5}] {}".format(
    #             reach.is_reachable_node(node), reach.is_safe_node(node),
    #             reach.state.resources.get(node.resource(), 0) > 0 if isinstance(node, ResourceNode) else "",
    #             game.world_list.node_name(node)))

    collected_indices = {
        resource
        for resource, quantity in reach.state.resources.items()
        if quantity > 0 and isinstance(resource, PickupIndex)
    }
    collected_events = {
        resource
        for resource, quantity in reach.state.resources.items()
        if quantity > 0 and resource.resource_type == ResourceType.EVENT
    }
    assert list(collectable_resource_nodes(reach.nodes, reach)) == []
    assert sorted(collected_indices) == expected_pickups
    assert sorted(collected_events, key=lambda it: it.index) == expected_events


@pytest.mark.parametrize("has_translator", [False, True])
def test_basic_search_with_translator_gate(has_translator: bool, echoes_resource_database):
    # Setup
    scan_visor = echoes_resource_database.get_item(10)

    node_a = GenericNode("Node A", True, None, 0)
    node_b = GenericNode("Node B", True, None, 1)
    node_c = GenericNode("Node C", True, None, 2)
    translator_node = TranslatorGateNode("Translator Gate", True, None, 3, TranslatorGate(1), scan_visor)

    world_list = WorldList([
        World("Test World", "Test Dark World", 1, [
            Area("Test Area A", False, 10, 0, True, [node_a, node_b, node_c, translator_node],
                 {
                     node_a: {
                         node_b: Requirement.trivial(),
                         translator_node: Requirement.trivial(),
                     },
                     node_b: {
                         node_a: Requirement.trivial(),
                     },
                     node_c: {
                         translator_node: Requirement.trivial(),
                     },
                     translator_node: {
                         node_a: Requirement.trivial(),
                         node_c: Requirement.trivial(),
                     },
                 }
                 )
        ])
    ])
    game_specific = EchoesGameSpecific(energy_per_tank=100, safe_zone_heal_per_second=1, beam_configurations=(),
                                       dangerous_energy_tank=False)
    game = GameDescription(RandovaniaGame.PRIME2, DockWeaknessDatabase([], [], [], []),
                           echoes_resource_database, game_specific, Requirement.impossible(),
                           None, {}, world_list)

    patches = game.create_game_patches()
    patches = patches.assign_gate_assignment({
        TranslatorGate(1): scan_visor
    })
    initial_state = State({scan_visor: 1 if has_translator else 0}, (), 99,
                          node_a, patches, None, echoes_resource_database, game.world_list)

    # Run
    reach = reach_with_all_safe_resources(game, initial_state)

    # Assert
    if has_translator:
        assert set(reach.safe_nodes) == {node_a, node_b, translator_node, node_c}
    else:
        assert set(reach.safe_nodes) == {node_a, node_b}


@pytest.mark.parametrize(["minimal_logic", "nodes", "safe_nodes"], [
    (False, 44, 5),
    (True, 1000, 1000),
])
def test_reach_size_from_start_echoes(echoes_game_description, default_layout_configuration, minimal_logic, nodes,
                                      safe_nodes):
    # Setup
    specific_levels = {
        trick.short_name: LayoutTrickLevel.HYPERMODE
        for trick in echoes_game_description.resource_database.trick
    }

    layout_configuration = dataclasses.replace(
        default_layout_configuration,
        trick_level=TrickLevelConfiguration(minimal_logic=minimal_logic,
                                            specific_levels=specific_levels if not minimal_logic else {},
                                            game=default_layout_configuration.game),
    )
    player_pool = generator.create_player_pool(Random(15000), layout_configuration, 0, 1)

    game, state = logic_bootstrap(layout_configuration, player_pool.game, player_pool.patches)

    # Run
    reach = GeneratorReach.reach_from_state(game, state)

    # Assert
    assert len(list(reach.nodes)) >= nodes
    assert len(list(reach.safe_nodes)) >= safe_nodes



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'debug_backend_window.ui'
##
## Created by: Qt User Interface Compiler version 5.14.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *


class Ui_DebugBackendWindow(object):
    def setupUi(self, DebugBackendWindow):
        if not DebugBackendWindow.objectName():
            DebugBackendWindow.setObjectName(u"DebugBackendWindow")
        DebugBackendWindow.resize(707, 322)
        self.central_widget = QWidget(DebugBackendWindow)
        self.central_widget.setObjectName(u"central_widget")
        self.central_widget.setMaximumSize(QSize(16777215, 16777215))
        self.central_widget.setLayoutDirection(Qt.LeftToRight)
        self.gridLayout = QGridLayout(self.central_widget)
        self.gridLayout.setSpacing(6)
        self.gridLayout.setContentsMargins(11, 11, 11, 11)
        self.gridLayout.setObjectName(u"gridLayout")
        self.reset_button = QPushButton(self.central_widget)
        self.reset_button.setObjectName(u"reset_button")

        self.gridLayout.addWidget(self.reset_button, 4, 3, 1, 1)

        self.messages_list = QListWidget(self.central_widget)
        self.messages_list.setObjectName(u"messages_list")

        self.gridLayout.addWidget(self.messages_list, 0, 2, 4, 2)

        self.current_status_label = QLabel(self.central_widget)
        self.current_status_label.setObjectName(u"current_status_label")

        self.gridLayout.addWidget(self.current_status_label, 0, 0, 1, 1)

        self.current_status_combo = QComboBox(self.central_widget)
        self.current_status_combo.setObjectName(u"current_status_combo")

        self.gridLayout.addWidget(self.current_status_combo, 0, 1, 1, 1)

        self.collect_location_button = QPushButton(self.central_widget)
        self.collect_location_button.setObjectName(u"collect_location_button")

        self.gridLayout.addWidget(self.collect_location_button, 1, 1, 1, 1)

        self.collect_location_combo = QComboBox(self.central_widget)
        self.collect_location_combo.setObjectName(u"collect_location_combo")

        self.gridLayout.addWidget(self.collect_location_combo, 1, 0, 1, 1)

        self.inventory_box = QGroupBox(self.central_widget)
        self.inventory_box.setObjectName(u"inventory_box")
        self.gridLayout_2 = QGridLayout(self.inventory_box)
        self.gridLayout_2.setSpacing(6)
        self.gridLayout_2.setContentsMargins(11, 11, 11, 11)
        self.gridLayout_2.setObjectName(u"gridLayout_2")
        self.inventory_label = QLabel(self.inventory_box)
        self.inventory_label.setObjectName(u"inventory_label")

        self.gridLayout_2.addWidget(self.inventory_label, 0, 0, 1, 1)


        self.gridLayout.addWidget(self.inventory_box, 2, 0, 2, 2)

        DebugBackendWindow.setCentralWidget(self.central_widget)
        self.menuBar = QMenuBar(DebugBackendWindow)
        self.menuBar.setObjectName(u"menuBar")
        self.menuBar.setGeometry(QRect(0, 0, 707, 21))
        DebugBackendWindow.setMenuBar(self.menuBar)

        self.retranslateUi(DebugBackendWindow)

        QMetaObject.connectSlotsByName(DebugBackendWindow)
    # setupUi

    def retranslateUi(self, DebugBackendWindow):
        DebugBackendWindow.setWindowTitle(QCoreApplication.translate("DebugBackendWindow", u"Debug Backend", None))
        self.reset_button.setText(QCoreApplication.translate("DebugBackendWindow", u"Reset", None))
        self.current_status_label.setText(QCoreApplication.translate("DebugBackendWindow", u"Current Status", None))
        self.collect_location_button.setText(QCoreApplication.translate("DebugBackendWindow", u"Collect Location", None))
        self.inventory_box.setTitle(QCoreApplication.translate("DebugBackendWindow", u"Inventory", None))
        self.inventory_label.setText("")
    # retranslateUi




import collections

import pytest
from PySide2 import QtWidgets, QtCore
from mock import MagicMock, AsyncMock, call

from randovania.game_description.resources.pickup_index import PickupIndex
from randovania.gui.seed_details_window import SeedDetailsWindow
from randovania.layout.layout_description import LayoutDescription


@pytest.mark.asyncio
async def test_export_iso(skip_qtbot, mocker):
    # Setup
    mock_execute_dialog = mocker.patch("randovania.gui.lib.async_dialog.execute_dialog", new_callable=AsyncMock,
                                       return_value=QtWidgets.QDialog.Accepted)

    options = MagicMock()
    options.output_directory = None

    window = SeedDetailsWindow(None, options)
    window.layout_description = MagicMock()
    window._player_names = {}
    window.run_in_background_async = AsyncMock()

    # Run
    await window._export_iso()

    # Assert
    mock_execute_dialog.assert_awaited_once()
    window.run_in_background_async.assert_awaited_once()


def test_update_layout_description_no_spoiler(skip_qtbot, mocker):
    # Setup
    mock_describer = mocker.patch("randovania.gui.lib.preset_describer.describe", return_value=["a", "b", "c", "d"])
    mock_merge = mocker.patch("randovania.gui.lib.preset_describer.merge_categories", return_value="<description>")

    options = MagicMock()
    description = MagicMock()
    description.permalink.player_count = 1
    description.permalink.as_base64_str = "<permalink>"
    description.permalink.spoiler = False

    window = SeedDetailsWindow(None, options)
    skip_qtbot.addWidget(window)

    # Run
    window.update_layout_description(description)

    # Assert
    mock_describer.assert_called_once_with(description.permalink.get_preset.return_value)
    mock_merge.assert_has_calls([
        call(["a", "c"]),
        call(["b", "d"]),
    ])


def test_update_layout_description_actual_seed(skip_qtbot, test_files_dir):
    description = LayoutDescription.from_file(test_files_dir.joinpath("log_files", "seed_a.rdvgame"))

    # Run
    window = SeedDetailsWindow(None, MagicMock())
    skip_qtbot.addWidget(window)
    window.update_layout_description(description)

    # Assert
    assert len(window.pickup_spoiler_buttons) == 119
    assert window.pickup_spoiler_show_all_button.text() == "Show All"
    skip_qtbot.mouseClick(window.pickup_spoiler_show_all_button, QtCore.Qt.LeftButton)
    assert window.pickup_spoiler_show_all_button.text() == "Hide All"


@pytest.mark.asyncio
async def test_show_dialog_for_prime3_layout(skip_qtbot, mocker, corruption_game_description):
    mock_execute_dialog = mocker.patch("randovania.gui.lib.async_dialog.execute_dialog", new_callable=AsyncMock)
    mock_clipboard: MagicMock = mocker.patch("PySide2.QtWidgets.QApplication.clipboard")

    window = SeedDetailsWindow(None, MagicMock())
    window.player_index_combo.addItem("Current", 0)
    skip_qtbot.addWidget(window)

    collections.namedtuple("MockPickup", ["name"])
    target = MagicMock()
    target.pickup.name = "Boost Ball"

    patches = corruption_game_description.create_game_patches()
    for i in range(100):
        patches.pickup_assignment[PickupIndex(i)] = target

    window.layout_description = MagicMock()
    window.layout_description.all_patches = {0: patches}

    # Run
    await window._show_dialog_for_prime3_layout()

    # Assert
    mock_execute_dialog.assert_awaited_once()
    mock_clipboard.return_value.setText.assert_called_once()



import dataclasses
from typing import NamedTuple, List, Dict, Optional, Iterator, Tuple

from randovania.game_description.node import Node, DockNode, PickupNode
from randovania.game_description.requirements import Requirement
from randovania.game_description.resources.pickup_index import PickupIndex


@dataclasses.dataclass(frozen=True)
class Area:
    name: str
    in_dark_aether: bool
    area_asset_id: int
    default_node_index: Optional[int]
    valid_starting_location: bool
    nodes: List[Node]
    connections: Dict[Node, Dict[Node, Requirement]]

    def __post_init__(self):
        object.__setattr__(self, "__cached_node_with_dock_index", {})

    def __repr__(self):
        return "Area[{}]".format(self.name)

    def __hash__(self):
        return self.area_asset_id

    def node_with_dock_index(self, dock_index: int) -> DockNode:
        cache: Dict[int, int] = object.__getattribute__(self, "__cached_node_with_dock_index")
        if dock_index in cache:
            return self.nodes[cache[dock_index]]

        for i, node in enumerate(self.nodes):
            if isinstance(node, DockNode) and node.dock_index == dock_index:
                cache[dock_index] = i
                return node
        raise IndexError("No DockNode found with dock_index {} in {}".format(
            dock_index, self.name))

    def node_with_name(self, node_name: str) -> Optional[Node]:
        """
        Searches this area for a node with the given name.
        :param node_name:
        :return: None, if not node is found
        """

        for node in self.nodes:
            if node.name == node_name:
                return node

        return None

    @property
    def all_connections(self) -> Iterator[Tuple[Node, Node, Requirement]]:
        """
        Iterates over all paths there are in this area.
        :return: source, target and the requirements for it
        """
        for source in self.nodes:
            for target, requirements in self.connections[source].items():
                yield source, target, requirements

    @property
    def pickup_indices(self) -> Iterator[PickupIndex]:
        for node in self.nodes:
            if isinstance(node, PickupNode):
                yield node.pickup_index

    @property
    def major_pickup_indices(self) -> Iterator[PickupIndex]:
        for node in self.nodes:
            if isinstance(node, PickupNode) and node.major_location:
                yield node.pickup_index

    def remove_node(self, node: Node):
        self.nodes.remove(node)
        cache: Dict[int, int] = object.__getattribute__(self, "__cached_node_with_dock_index")
        cache.clear()



import asyncio
import json
import platform
import re
import shutil
from asyncio import StreamWriter, StreamReader
try:
    from asyncio.exceptions import IncompleteReadError
except ImportError:
    from asyncio.streams import IncompleteReadError
from pathlib import Path
from typing import Callable, List, Union, Optional

from randovania import get_data_path
from randovania.game_description.echoes_game_specific import EchoesGameSpecific
from randovania.games.prime import patcher_file, dol_patcher
from randovania.interface_common import status_update_lib
from randovania.interface_common.cosmetic_patches import CosmeticPatches
from randovania.interface_common.echoes_user_preferences import EchoesUserPreferences
from randovania.interface_common.game_workdir import validate_game_files_path
from randovania.interface_common.players_configuration import PlayersConfiguration
from randovania.interface_common.status_update_lib import ProgressUpdateCallable
from randovania.layout.layout_description import LayoutDescription

IO_LOOP: Optional[asyncio.AbstractEventLoop] = None


CURRENT_PATCH_VERSION = 1


def _patch_version_file(game_root: Path) -> Path:
    return game_root.joinpath("randovania_patch_version.txt")


def get_patch_version(game_root: Path) -> int:
    file = _patch_version_file(game_root)
    if file.exists():
        return int(file.read_text("utf-8"))
    else:
        return 0


def write_patch_version(game_root: Path, version: int):
    _patch_version_file(game_root).write_text(str(version))


def _get_randomizer_folder() -> Path:
    return get_data_path().joinpath("ClarisPrimeRandomizer")


def _get_randomizer_path() -> Path:
    return _get_randomizer_folder().joinpath("Randomizer.exe")


def _get_menu_mod_path() -> Path:
    return get_data_path().joinpath("ClarisEchoesMenu", "EchoesMenu.exe")


def _is_windows() -> bool:
    return platform.system() == "Windows"


async def _write_data(stream: StreamWriter, data: str):
    stream.write(data.encode("UTF-8"))
    stream.close()


async def _read_data(stream: StreamReader, read_callback: Callable[[str], None]):
    while True:
        try:
            line = await stream.readuntil(b"\r")
        except IncompleteReadError as incomplete:
            line = incomplete.partial
        if line:
            try:
                decoded = line.decode()
            except UnicodeDecodeError:
                decoded = line.decode("latin1")
            for x in re.split(r"[\r\n]", decoded.strip()):
                if x:
                    read_callback(x)
        else:
            break


async def _process_command_async(args: List[str], input_data: str, read_callback: Callable[[str], None]):
    process = await asyncio.create_subprocess_exec(*args,
                                                   stdin=asyncio.subprocess.PIPE,
                                                   stdout=asyncio.subprocess.PIPE,
                                                   stderr=asyncio.subprocess.STDOUT)

    await asyncio.wait([
        _write_data(process.stdin, input_data),
        _read_data(process.stdout, read_callback),
    ])
    await process.wait()


def _process_command(args: List[str], input_data: str, read_callback: Callable[[str], None]):
    work = _process_command_async(args, input_data, read_callback)

    if IO_LOOP is None:
        if _is_windows():
            asyncio.set_event_loop_policy(asyncio.WindowsProactorEventLoopPolicy())
        asyncio.run(work)
    else:
        asyncio.run_coroutine_threadsafe(work, IO_LOOP).result()


def _run_with_args(args: List[Union[str, Path]],
                   input_data: str,
                   finish_string: str,
                   status_update: Callable[[str], None]):
    finished_updates = False

    new_args = [str(arg) for arg in args]
    if not _is_windows():
        new_args.insert(0, "mono")
    print("Invoking external tool with: ", new_args)

    def read_callback(line: str):
        nonlocal finished_updates
        print(line)
        if not finished_updates:
            status_update(line)
            finished_updates = line == finish_string

    _process_command(new_args, input_data, read_callback)

    if not finished_updates:
        raise RuntimeError("External tool did not send '{}'. Did something happen?".format(finish_string))


def _base_args(game_root: Path,
               ) -> List[Union[str, Path]]:
    game_files = game_root / "files"
    validate_game_files_path(game_files)

    return [
        _get_randomizer_path(),
        game_root,
        # "-test",
    ]


def _ensure_no_menu_mod(
        game_root: Path,
        backup_files_path: Optional[Path],
        status_update: Callable[[str], None],
):
    """
    Ensures the given game_root has no menu mod, copying paks from the backup path if needed.
    :param game_root:
    :param backup_files_path:
    :param status_update:
    :return:
    """
    files_folder = game_root.joinpath("files")
    menu_mod_txt = files_folder.joinpath("menu_mod.txt")

    if menu_mod_txt.is_file():
        if backup_files_path is None:
            raise RuntimeError("Game at '{}' has Menu Mod, but no backup path given to restore".format(game_root))

        pak_folder = backup_files_path.joinpath("mp2_paks")
        if pak_folder.is_dir():
            for pak in pak_folder.glob("**/*.pak"):
                relative = pak.relative_to(pak_folder)
                status_update("Restoring {} from backup".format(relative))
                shutil.copy(pak, files_folder.joinpath(relative))

            menu_mod_txt.unlink()


_ECHOES_PAKS = tuple(["MiscData.pak"] + ["Metroid{}.pak".format(i) for i in range(1, 6)])


def _create_pak_backups(
        game_root: Path,
        backup_files_path: Path,
        status_update: Callable[[str], None],
):
    pak_folder = backup_files_path.joinpath("mp2_paks")
    pak_folder.mkdir(parents=True, exist_ok=True)

    files_folder = game_root.joinpath("files")
    for pak in _ECHOES_PAKS:
        target_file = pak_folder.joinpath(pak)
        if not target_file.exists():
            status_update("Backing up {}".format(pak))
            shutil.copy(files_folder.joinpath(pak), target_file)


def _add_menu_mod_to_files(
        game_root: Path,
        status_update: Callable[[str], None],
):
    files_folder = game_root.joinpath("files")
    _run_with_args(
        [_get_menu_mod_path(), files_folder],
        "",
        "Done!",
        status_update
    )
    files_folder.joinpath("menu_mod.txt").write_bytes(b"")


def apply_layout(description: LayoutDescription,
                 players_config: PlayersConfiguration,
                 cosmetic_patches: CosmeticPatches,
                 backup_files_path: Optional[Path],
                 progress_update: ProgressUpdateCallable,
                 game_root: Path,
                 ):
    """
    Applies the modifications listed in the given LayoutDescription to the game in game_root.
    :param description:
    :param players_config:
    :param cosmetic_patches:
    :param game_root:
    :param backup_files_path: Path to use as pak backup, to remove/add menu mod.
    :param progress_update:
    :return:
    """
    description.save_to_file(game_root.joinpath("files", f"randovania.{description.file_extension()}"))

    apply_patcher_file(game_root, backup_files_path,
                       patcher_file.create_patcher_file(description, players_config, cosmetic_patches),
                       description.all_patches[players_config.player_index].game_specific,
                       progress_update)


def apply_patcher_file(game_root: Path,
                       backup_files_path: Optional[Path],
                       patcher_data: dict,
                       game_specific: EchoesGameSpecific,
                       progress_update: ProgressUpdateCallable,
                       ):
    """
    Applies the modifications listed in the given patcher_data to the game in game_root.
    :param game_root:
    :param backup_files_path:
    :param patcher_data:
    :param game_specific:
    :param progress_update:
    :return:
    """
    menu_mod = patcher_data["menu_mod"]
    user_preferences = EchoesUserPreferences.from_json_dict(patcher_data["user_preferences"])
    default_items = patcher_data["default_items"]

    status_update = status_update_lib.create_progress_update_from_successive_messages(
        progress_update, 400 if menu_mod else 100)

    last_version = get_patch_version(game_root)
    if last_version > CURRENT_PATCH_VERSION:
        raise RuntimeError(f"Game at {game_root} was last patched with version {last_version}, "
                           f"which is above supported version {CURRENT_PATCH_VERSION}. "
                           f"\nPlease press 'Delete internal copy'.")

    _ensure_no_menu_mod(game_root, backup_files_path, status_update)
    if backup_files_path is not None:
        _create_pak_backups(game_root, backup_files_path, status_update)

    _run_with_args(_base_args(game_root),
                   json.dumps(patcher_data),
                   "Randomized!",
                   status_update)
    dol_patcher.apply_patches(game_root, game_specific, user_preferences, default_items)
    write_patch_version(game_root, CURRENT_PATCH_VERSION)

    if menu_mod:
        _add_menu_mod_to_files(game_root, status_update)



from PySide2 import QtCore

from randovania.gui.preset_settings.item_configuration_widget import ItemConfigurationWidget
from randovania.layout.major_item_state import MajorItemState


def test_state_no_changes(skip_qtbot, echoes_item_database, echoes_resource_database):
    item = [item for item in echoes_item_database.major_items.values() if not item.required][0]
    state = MajorItemState(
        include_copy_in_original_location=False,
        num_shuffled_pickups=1,
        num_included_in_starting_items=0,
        included_ammo=(),
    )

    # Run
    popup = ItemConfigurationWidget(None, item, state, echoes_resource_database)
    skip_qtbot.addWidget(popup)

    # Assert
    assert popup.state == state


def test_state_change_to_starting(skip_qtbot, echoes_item_database, echoes_resource_database):
    item = [item for item in echoes_item_database.major_items.values() if not item.required][0]
    state = MajorItemState(
        include_copy_in_original_location=False,
        num_shuffled_pickups=1,
        num_included_in_starting_items=0,
        included_ammo=(),
    )

    # Run
    popup = ItemConfigurationWidget(None, item, state, echoes_resource_database)
    skip_qtbot.addWidget(popup)
    skip_qtbot.mouseClick(popup.starting_radio, QtCore.Qt.LeftButton)

    # Assert
    assert popup.state == MajorItemState(
        include_copy_in_original_location=False,
        num_shuffled_pickups=0,
        num_included_in_starting_items=1,
        included_ammo=(),
    )


def test_state_required(skip_qtbot, echoes_item_database, echoes_resource_database):
    item = [item for item in echoes_item_database.major_items.values() if item.required][0]
    state = MajorItemState(
        include_copy_in_original_location=False,
        num_shuffled_pickups=1,
        num_included_in_starting_items=0,
        included_ammo=(),
    )

    # Run
    popup = ItemConfigurationWidget(None, item, state, echoes_resource_database)
    skip_qtbot.addWidget(popup)

    # Assert
    assert popup.state == MajorItemState(
        include_copy_in_original_location=False,
        num_shuffled_pickups=0,
        num_included_in_starting_items=1,
        included_ammo=(),
    )



from typing import Iterable, Dict

from PySide2 import QtWidgets
from PySide2.QtCore import Qt

from randovania.game_description.item.major_item import MajorItem
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.major_item_state import MajorItemState
from randovania.layout.preset import Preset


def _state_has_item(state: MajorItemState) -> bool:
    return (state.num_shuffled_pickups > 0
            or state.num_included_in_starting_items > 0
            or state.include_copy_in_original_location)


class ProgressiveItemWidget(QtWidgets.QCheckBox):
    def __init__(self, parent: QtWidgets.QWidget, editor: PresetEditor,
                 progressive_item: MajorItem, non_progressive_items: Iterable[MajorItem]):
        super().__init__(parent)
        self._editor = editor
        self.progressive_item = progressive_item
        self.non_progressive_items = list(non_progressive_items)
        self.setTristate(True)
        self.clicked.connect(self.change_progressive)

    def on_preset_changed(self, preset: Preset, elements: Dict[MajorItem, QtWidgets.QWidget]):
        major_configuration = preset.configuration.major_items_configuration

        has_progressive = _state_has_item(major_configuration.items_state[self.progressive_item])
        has_non_progressive = any(_state_has_item(major_configuration.items_state[item])
                                  for item in self.non_progressive_items)

        if has_non_progressive:
            new_state = Qt.PartiallyChecked if has_progressive else Qt.Unchecked
        else:
            new_state = Qt.Checked
        self.setCheckState(new_state)

        for item in self.non_progressive_items:
            elements[item].setVisible(has_non_progressive)
        elements[self.progressive_item].setVisible(has_progressive)

    def change_progressive(self, is_progressive: bool):
        with self._editor as editor:
            if is_progressive:
                non_progressive_state = MajorItemState()
                progressive_state = MajorItemState(num_shuffled_pickups=len(self.non_progressive_items))
            else:
                non_progressive_state = MajorItemState(num_shuffled_pickups=1)
                progressive_state = MajorItemState()

            new_states = {self.progressive_item: progressive_state}
            for item in self.non_progressive_items:
                new_states[item] = non_progressive_state

            editor.major_items_configuration = editor.major_items_configuration.replace_states(new_states)



from PySide2.QtWidgets import QComboBox

from randovania.gui.generated.preset_echoes_patches_ui import Ui_PresetEchoesPatches
from randovania.gui.preset_settings.preset_tab import PresetTab
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.pickup_model import PickupModelStyle, PickupModelDataSource
from randovania.layout.preset import Preset


class PresetEchoesPatches(PresetTab, Ui_PresetEchoesPatches):
    _editor: PresetEditor

    def __init__(self, editor: PresetEditor):
        super().__init__()
        self.setupUi(self)
        self._editor = editor

        # Item Data
        for i, value in enumerate(PickupModelStyle):
            self.pickup_model_combo.setItemData(i, value)

        for i, value in enumerate(PickupModelDataSource):
            self.pickup_data_source_combo.setItemData(i, value)

        self.include_menu_mod_label.setText(self.include_menu_mod_label.text().replace("color:#0000ff;", ""))

        # TODO: implement the LOCATION data source
        self.pickup_data_source_combo.removeItem(self.pickup_data_source_combo.findData(PickupModelDataSource.LOCATION))

        # Signals
        self.warp_to_start_check.stateChanged.connect(self._persist_option_then_notify("warp_to_start"))
        self.include_menu_mod_check.stateChanged.connect(self._persist_option_then_notify("include_menu_mod"))
        self.skip_final_bosses_check.stateChanged.connect(self._persist_option_then_notify("skip_final_bosses"))

        self.pickup_model_combo.currentIndexChanged.connect(self._persist_enum(self.pickup_model_combo,
                                                                               "pickup_model_style"))
        self.pickup_data_source_combo.currentIndexChanged.connect(self._persist_enum(self.pickup_data_source_combo,
                                                                                     "pickup_model_data_source"))

    @property
    def uses_patches_tab(self) -> bool:
        return True

    def _persist_option_then_notify(self, attribute_name: str):
        def persist(value: int):
            with self._editor as options:
                setattr(options, attribute_name, bool(value))

        return persist

    def _persist_enum(self, combo: QComboBox, attribute_name: str):
        def persist(index: int):
            with self._editor as options:
                options.set_configuration_field(attribute_name, combo.itemData(index))

        return persist

    def on_preset_changed(self, preset: Preset):
        config = preset.configuration
        self.warp_to_start_check.setChecked(config.warp_to_start)
        self.include_menu_mod_check.setChecked(config.menu_mod)
        self.skip_final_bosses_check.setChecked(config.skip_final_bosses)

        self.pickup_model_combo.setCurrentIndex(self.pickup_model_combo.findData(config.pickup_model_style))
        self.pickup_data_source_combo.setCurrentIndex(
            self.pickup_data_source_combo.findData(config.pickup_model_data_source))
        self.pickup_data_source_combo.setEnabled(config.pickup_model_style != PickupModelStyle.ALL_VISIBLE)



import dataclasses
from functools import partial

from PySide2.QtWidgets import QDialog, QWidget, QCheckBox, QSlider, QLabel

from randovania.gui.generated.echoes_user_preferences_dialog_ui import Ui_EchoesUserPreferencesDialog
from randovania.interface_common.cosmetic_patches import CosmeticPatches
from randovania.interface_common.echoes_user_preferences import EchoesUserPreferences, SoundMode


def update_label_with_slider(label: QLabel, slider: QSlider):
    if label.display_as_percentage:
        min_value = slider.minimum()
        percentage = (slider.value() - min_value) / (slider.maximum() - min_value)
        label.setText(f"{percentage * 100: 3.0f}%")
    else:
        label.setText(str(slider.value()))


class EchoesUserPreferencesDialog(QDialog, Ui_EchoesUserPreferencesDialog):
    preferences: EchoesUserPreferences
    cosmetic_patches: CosmeticPatches

    def __init__(self, parent: QWidget, current: CosmeticPatches):
        super().__init__(parent)
        self.setupUi(self)
        self.cosmetic_patches = current

        self.field_to_slider_mapping = {
            "screen_brightness": self.screen_brightness_slider,
            "screen_x_offset": self.screen_x_offset_slider,
            "screen_y_offset": self.screen_y_offset_slider,
            "screen_stretch": self.screen_stretch_slider,
            "sfx_volume": self.sfx_volume_slider,
            "music_volume": self.music_volume_slider,
            "hud_alpha": self.hud_alpha_slider,
            "helmet_alpha": self.helmet_alpha_slider,
        }
        self.field_to_check_mapping = {
            "hud_lag": self.hud_lag_check,
            "invert_y_axis": self.invert_y_axis_check,
            "rumble": self.rumble_check,
        }

        for sound_mode in SoundMode:
            self.sound_mode_combo.addItem(sound_mode.name, sound_mode)

        self.on_new_cosmetic_patches(current)

        self.remove_hud_popup_check.stateChanged.connect(self._persist_option_then_notify("disable_hud_popup"))
        self.faster_credits_check.stateChanged.connect(self._persist_option_then_notify("speed_up_credits"))
        self.open_map_check.stateChanged.connect(self._persist_option_then_notify("open_map"))
        self.pickup_markers_check.stateChanged.connect(self._persist_option_then_notify("pickup_markers"))

        self.sound_mode_combo.currentIndexChanged.connect(self._on_sound_mode_update)

        for field_name, slider in self.field_to_slider_mapping.items():
            slider.valueChanged.connect(partial(self._on_slider_update, slider, field_name))

        for field_name, check in self.field_to_check_mapping.items():
            check.stateChanged.connect(partial(self._on_check_update, check, field_name))

        self.accept_button.clicked.connect(self.accept)
        self.cancel_button.clicked.connect(self.reject)
        self.reset_button.clicked.connect(self.reset)

    def on_new_cosmetic_patches(self, patches: CosmeticPatches):
        self.remove_hud_popup_check.setChecked(patches.disable_hud_popup)
        self.faster_credits_check.setChecked(patches.speed_up_credits)
        self.open_map_check.setChecked(patches.open_map)
        self.pickup_markers_check.setChecked(patches.pickup_markers)
        self.on_new_user_preferences(patches.user_preferences)

    def on_new_user_preferences(self, user_preferences: EchoesUserPreferences):
        self.sound_mode_combo.setCurrentIndex(self.sound_mode_combo.findData(user_preferences.sound_mode))

        for field in dataclasses.fields(user_preferences):
            if field.name in self.field_to_slider_mapping:
                slider = self.field_to_slider_mapping[field.name]
                slider.setMinimum(field.metadata["min"])
                slider.setMaximum(field.metadata["max"])
                slider.setValue(getattr(user_preferences, field.name))

                value_label: QLabel = getattr(self, f"{field.name}_value_label")
                value_label.display_as_percentage = field.metadata["display_as_percentage"]
                update_label_with_slider(value_label, slider)

            elif field.name in self.field_to_check_mapping:
                check = self.field_to_check_mapping[field.name]
                check.setChecked(getattr(user_preferences, field.name))

    def _persist_option_then_notify(self, attribute_name: str):
        def persist(value: int):
            self.cosmetic_patches = dataclasses.replace(
                self.cosmetic_patches,
                **{attribute_name: bool(value)}
            )

        return persist

    @property
    def preferences(self):
        return self.cosmetic_patches.user_preferences

    @preferences.setter
    def preferences(self, value):
        self.cosmetic_patches = dataclasses.replace(
            self.cosmetic_patches,
            user_preferences=value,
        )

    def _on_sound_mode_update(self):
        self.preferences = dataclasses.replace(
            self.preferences,
            sound_mode=self.sound_mode_combo.currentData()
        )

    def _on_slider_update(self, slider: QSlider, field_name: str, _):
        self.preferences = dataclasses.replace(
            self.preferences,
            **{field_name: slider.value()}
        )
        update_label_with_slider(getattr(self, f"{field_name}_value_label"), slider)

    def _on_check_update(self, check: QCheckBox, field_name: str, _):
        self.preferences = dataclasses.replace(
            self.preferences,
            **{field_name: check.isChecked()}
        )

    def reset(self):
        self.on_new_cosmetic_patches(CosmeticPatches())



import pytest
from PySide2 import QtWidgets
from mock import MagicMock, AsyncMock

from randovania.games.game import RandovaniaGame
from randovania.gui.generate_seed_tab import GenerateSeedTab
from randovania.gui.generated.main_window_ui import Ui_MainWindow


@pytest.fixture(name="tab")
def _tab(skip_qtbot):
    class Window(QtWidgets.QMainWindow, Ui_MainWindow):
        pass

    window = Window()
    window.setupUi(window)

    return GenerateSeedTab(window, MagicMock(), MagicMock())


def test_add_new_preset(tab, preset_manager):
    tab.setup_ui()
    tab._window_manager.preset_manager.add_new_preset.return_value = True

    # Run
    tab._add_new_preset(preset_manager.default_preset)

    # Assert
    tab._window_manager.preset_manager.add_new_preset.assert_called_once_with(preset_manager.default_preset)
    assert not tab._action_delete.isEnabled()


@pytest.mark.parametrize("has_existing_window", [False, True])
@pytest.mark.asyncio
async def test_on_customize_button(tab, mocker, has_existing_window):
    mock_settings_window = mocker.patch("randovania.gui.generate_seed_tab.LogicSettingsWindow")
    mock_execute_dialog = mocker.patch("randovania.gui.lib.async_dialog.execute_dialog", new_callable=AsyncMock)
    mock_execute_dialog.return_value = QtWidgets.QDialog.Accepted
    tab._add_new_preset = MagicMock()
    tab._logic_settings_window = MagicMock() if has_existing_window else None

    # Run
    await tab._on_customize_button()

    # Assert
    if has_existing_window:
        tab._logic_settings_window.raise_.assert_called_once_with()
        mock_settings_window.assert_not_called()
        mock_execute_dialog.assert_not_awaited()
        tab._add_new_preset.assert_not_called()
    else:
        mock_settings_window.assert_called_once()
        mock_execute_dialog.assert_awaited_once_with(mock_settings_window.return_value)
        tab._add_new_preset.assert_called_once()


@pytest.mark.parametrize("game", [RandovaniaGame.PRIME2, RandovaniaGame.PRIME3])
def test_on_options_changed_select_preset(tab, preset_manager, game):
    preset = preset_manager.default_preset_for_game(game)

    tab._window_manager.preset_manager = preset_manager
    tab.setup_ui()

    tab._options.selected_preset_name = preset.name

    # Run
    tab.on_options_changed(tab._options)

    # Assert
    assert tab._current_preset_data == preset


def test_select_game(tab, preset_manager):
    tab._window_manager.preset_manager = preset_manager
    tab.setup_ui()

    tab.select_game(RandovaniaGame.PRIME3)
    assert tab._current_preset_data.game == RandovaniaGame.PRIME3

    tab.select_game(RandovaniaGame.PRIME2)
    assert tab._current_preset_data.game == RandovaniaGame.PRIME2



from unittest.mock import MagicMock

import pytest

from randovania.game_description.resources.resource_type import ResourceType
from randovania.layout.elevators import LayoutElevators
from randovania.resolver import bootstrap


@pytest.mark.parametrize("elevators", [LayoutElevators.VANILLA, LayoutElevators.TWO_WAY_RANDOMIZED])
def test_misc_resources_for_configuration(echoes_resource_database,
                                          elevators: LayoutElevators,
                                          ):
    # Setup
    configuration = MagicMock()
    configuration.elevators = elevators
    gfmc_resource = echoes_resource_database.get_by_type_and_index(ResourceType.MISC, 16)
    torvus_resource = echoes_resource_database.get_by_type_and_index(ResourceType.MISC, 17)
    great_resource = echoes_resource_database.get_by_type_and_index(ResourceType.MISC, 18)

    # Run
    result = bootstrap.misc_resources_for_configuration(configuration, echoes_resource_database)
    relevant_tricks = {
        trick: result[trick]
        for trick in [gfmc_resource, torvus_resource, great_resource]
    }

    # Assert
    assert relevant_tricks == {
        gfmc_resource: 0,
        torvus_resource: 0,
        great_resource: 0 if elevators != LayoutElevators.VANILLA else 1,
    }


def test_logic_bootstrap(default_preset, echoes_game_description):
    new_game, state = bootstrap.logic_bootstrap(default_preset.configuration, echoes_game_description,
                                                echoes_game_description.create_game_patches())

    



import pytest
from mock import MagicMock, AsyncMock, call

from randovania.game_connection.connection_backend import MemoryOperation, MemoryOperationException
from randovania.game_connection.connection_base import GameConnectionStatus
from randovania.game_connection.nintendont_backend import NintendontBackend, SocketHolder


@pytest.fixture(name="backend")
def nintendont_backend():
    backend = NintendontBackend("localhost")
    return backend


@pytest.mark.asyncio
async def test_perform_memory_operations_success(backend: NintendontBackend):
    backend._socket = MagicMock()
    backend._socket.max_input = 120
    backend._socket.max_output = 100
    backend._socket.max_addresses = 8
    backend._socket.writer.drain = AsyncMock()
    backend._socket.reader.read = AsyncMock(side_effect=[
        b"\x03" + b"A" * 50 + b"B" * 30,
        b"\x01" + b"C" * 60,
    ])
    ops = {
        MemoryOperation(0x1000, read_byte_count=50): b"A" * 50,
        MemoryOperation(0x1000, offset=10, read_byte_count=30, write_bytes=b"1" * 30): b"B" * 30,
        MemoryOperation(0x1000, read_byte_count=60): b"C" * 60,
    }

    # Run
    result = await backend._perform_memory_operations(list(ops.keys()))

    # Assert
    backend._socket.writer.drain.assert_has_awaits([call(), call()])
    backend._socket.writer.write.assert_has_calls([
        call(b'\x00\x02\x01\x01\x00\x00\x10\x00' + b'\x80\x32' + b'\xd0\x1e\x00\n' + (b"1" * 30)),
        call(b'\x00\x01\x01\x01\x00\x00\x10\x00\x80\x3c'),
    ])
    assert result == ops
    backend._socket.reader.read.assert_has_awaits([call(1024), call(1024)])


@pytest.mark.asyncio
async def test_perform_memory_operations_invalid(backend: NintendontBackend):
    backend._socket = MagicMock()
    backend._socket.max_input = 120
    backend._socket.max_output = 100
    backend._socket.max_addresses = 8
    backend._socket.writer.drain = AsyncMock()
    backend._socket.reader.read = AsyncMock(side_effect=[
        b"\x03" + b"A" * 50 + b"B" * 30,
    ])

    # Run
    with pytest.raises(MemoryOperationException):
        await backend._perform_memory_operations([
            MemoryOperation(0x1000, read_byte_count=50),
            MemoryOperation(0x2000, read_byte_count=10),
            MemoryOperation(0x2000, read_byte_count=10),
        ])

    # Assert
    backend._socket.writer.drain.assert_has_awaits([call()])
    backend._socket.writer.write.assert_has_calls([
        call(b'\x00\x03\x02\x01\x00\x00\x10\x00\x00\x00 \x00\x802\x81\n\x81\n'),
    ])
    backend._socket.reader.read.assert_has_awaits([call(1024)])


@pytest.mark.asyncio
async def test_perform_single_giant_memory_operation(backend: NintendontBackend):
    backend._socket = MagicMock()
    backend._socket.max_input = 120
    backend._socket.max_output = 100
    backend._socket.max_addresses = 8
    backend._socket.writer.drain = AsyncMock()
    backend._socket.reader.read = AsyncMock(side_effect=[
        b"\x01",
        b"\x01",
    ])

    # Run
    result = await backend._perform_single_memory_operations(
        MemoryOperation(0x1000, write_bytes=b"1" * 200),
    )

    # Assert
    backend._socket.writer.drain.assert_has_awaits([call(), call()])
    backend._socket.writer.write.assert_has_calls([
        call(b'\x00\x01\x01\x01\x00\x00\x10\x00' + b'\x40\x64' + (b"1" * 100)),
        call(b'\x00\x01\x01\x01\x00\x00\x10\x64' + b'\x40\x64' + (b"1" * 100)),
    ])
    assert result is None
    backend._socket.reader.read.assert_has_awaits([call(1024), call(1024)])


@pytest.mark.asyncio
async def test_connect(backend, mocker):
    reader, writer = MagicMock(), MagicMock()
    writer.drain = AsyncMock()
    reader.read = AsyncMock(return_value=b'\x00\x00\x00\x02\x00\x00\x00x\x00\x00\x00\xfa\x00\x00\x00\x03')
    mock_open = mocker.patch("asyncio.open_connection", new_callable=AsyncMock, return_value=(reader, writer))

    # Run
    await backend._connect()

    # Assert
    mock_open.assert_awaited_once_with("localhost", backend._port)
    writer.drain.assert_awaited_once_with()

    socket: SocketHolder = backend._socket
    assert socket.reader is reader
    assert socket.writer is writer
    assert socket.api_version == 2
    assert socket.max_input == 120
    assert socket.max_output == 250
    assert socket.max_addresses == 3


def test_current_status_disconnected(backend):
    backend._socket = None
    assert backend.current_status == GameConnectionStatus.Disconnected


def test_current_status_wrong_game(backend):
    backend._socket = True
    assert backend.current_status == GameConnectionStatus.UnknownGame


def test_current_status_not_in_game(backend):
    backend._socket = True
    backend.patches = True
    assert backend.current_status == GameConnectionStatus.TitleScreen


def test_current_status_tracker_only(backend):
    backend._socket = True
    backend.patches = True
    backend._world = True
    assert backend.current_status == GameConnectionStatus.TrackerOnly


def test_current_status_in_game(backend):
    backend._socket = True
    backend.patches = True
    backend._world = True
    backend.checking_for_collected_index = True
    assert backend.current_status == GameConnectionStatus.InGame



from distutils.version import StrictVersion
from enum import Enum
from logging.config import dictConfig

import flask
from flask_socketio import ConnectionRefusedError

import randovania
from randovania.server import game_session, user_session, database
from randovania.server.server_app import ServerApp


class ClientVersionCheck(Enum):
    STRICT = "strict"
    MATCH_MAJOR_MINOR = "match-major-minor"
    IGNORE = "ignore"


def check_client_version(version_checking: ClientVersionCheck, client_version: str, server_version: str):
    if version_checking == ClientVersionCheck.STRICT:
        if server_version != client_version:
            raise ConnectionRefusedError(f"Incompatible client version '{client_version}', "
                                         f"expected '{server_version}'")
    elif version_checking == ClientVersionCheck.MATCH_MAJOR_MINOR:
        server = StrictVersion(server_version.split(".dev")[0])
        client = StrictVersion(client_version.split(".dev")[0])
        if server.version[:2] != client.version[:2]:
            shorter_client = "{}.{}".format(*client.version[:2])
            shorter_server = "{}.{}".format(*server.version[:2])
            raise ConnectionRefusedError(f"Incompatible client version '{shorter_client}', "
                                         f"expected '{shorter_server}'")


def create_app():
    configuration = randovania.get_configuration()

    dictConfig({
        'version': 1,
        'formatters': {'default': {
            'format': '[%(asctime)s] %(levelname)s in %(module)s: %(message)s',
        }},
        'handlers': {'wsgi': {
            'class': 'logging.StreamHandler',
            'stream': 'ext://flask.logging.wsgi_errors_stream',
            'formatter': 'default'
        }},
        'root': {
            'level': 'INFO',
            'handlers': ['wsgi']
        }
    })

    app = flask.Flask(__name__)
    app.config['SECRET_KEY'] = configuration["server_config"]["secret_key"]
    app.config["GUEST_KEY"] = configuration["guest_secret"].encode("ascii") if "guest_secret" in configuration else None
    app.config["DISCORD_CLIENT_ID"] = configuration["discord_client_id"]
    app.config["DISCORD_CLIENT_SECRET"] = configuration["server_config"]["discord_client_secret"]
    app.config["DISCORD_REDIRECT_URI"] = "http://127.0.0.1:5000/callback/"  # Redirect URI.
    app.config["FERNET_KEY"] = configuration["server_config"]["fernet_key"].encode("ascii")
    version_checking = ClientVersionCheck(configuration["server_config"]["client_version_checking"])

    database.db.init(configuration["server_config"]['database_path'])
    database.db.connect(reuse_if_open=True)
    database.db.create_tables(database.all_classes)

    sio = ServerApp(app)
    app.sio = sio
    game_session.setup_app(sio)
    user_session.setup_app(sio)

    connected_clients = sio.metrics.info("connected_clients", "How many clients are connected right now.")
    connected_clients.set(0)

    @app.route("/")
    def index():
        return "ok"

    server_version = randovania.VERSION

    @sio.sio.server.on("connect")
    def connect(sid, environ):
        if "HTTP_X_RANDOVANIA_VERSION" not in environ:
            raise ConnectionRefusedError("unknown client version")

        client_app_version = environ["HTTP_X_RANDOVANIA_VERSION"]
        check_client_version(version_checking, client_app_version, server_version)
        connected_clients.inc()

        forwarded_for = environ.get('HTTP_X_FORWARDED_FOR')
        app.logger.info(f"Client at {environ['REMOTE_ADDR']} ({forwarded_for}) with "
                        f"version {client_app_version} connected.")

    @sio.sio.server.on("disconnect")
    def disconnect(sid):
        connected_clients.dec()
        sio_environ = sio.get_server().environ

        forwarded_for = sio_environ[sid].get('HTTP_X_FORWARDED_FOR')
        app.logger.info(f"Client at {sio_environ[sid]['REMOTE_ADDR']} ({forwarded_for}) disconnected.")

        session = sio.get_server().get_session(sid)
        if "user-id" in session:
            game_session.report_user_disconnected(sio, session["user-id"], app.logger)

    return app



from unittest.mock import MagicMock

from randovania.gui.lib import common_qt_lib


def test_get_network_client(skip_qtbot, qapp):
    qapp.network_client = MagicMock()

    assert common_qt_lib.get_network_client() is qapp.network_client


def test_get_game_connection(skip_qtbot, qapp):
    qapp.game_connection = MagicMock()

    assert common_qt_lib.get_game_connection() is qapp.game_connection



# -*- coding: utf-8 -*-

################################################################################
## Form generated from reading UI file 'preset_echoes_goal.ui'
##
## Created by: Qt User Interface Compiler version 5.14.2
##
## WARNING! All changes made in this file will be lost when recompiling UI file!
################################################################################

from PySide2.QtCore import (QCoreApplication, QDate, QDateTime, QMetaObject,
    QObject, QPoint, QRect, QSize, QTime, QUrl, Qt)
from PySide2.QtGui import (QBrush, QColor, QConicalGradient, QCursor, QFont,
    QFontDatabase, QIcon, QKeySequence, QLinearGradient, QPalette, QPainter,
    QPixmap, QRadialGradient)
from PySide2.QtWidgets import *


class Ui_PresetEchoesGoal(object):
    def setupUi(self, PresetEchoesGoal):
        if not PresetEchoesGoal.objectName():
            PresetEchoesGoal.setObjectName(u"PresetEchoesGoal")
        PresetEchoesGoal.resize(562, 459)
        self.centralWidget = QWidget(PresetEchoesGoal)
        self.centralWidget.setObjectName(u"centralWidget")
        self.centralWidget.setMaximumSize(QSize(16777215, 16777215))
        self.goal_layout = QVBoxLayout(self.centralWidget)
        self.goal_layout.setSpacing(6)
        self.goal_layout.setContentsMargins(11, 11, 11, 11)
        self.goal_layout.setObjectName(u"goal_layout")
        self.goal_layout.setContentsMargins(4, 8, 4, 0)
        self.skytemple_description = QLabel(self.centralWidget)
        self.skytemple_description.setObjectName(u"skytemple_description")
        self.skytemple_description.setWordWrap(True)

        self.goal_layout.addWidget(self.skytemple_description)

        self.skytemple_combo = QComboBox(self.centralWidget)
        self.skytemple_combo.addItem("")
        self.skytemple_combo.addItem("")
        self.skytemple_combo.addItem("")
        self.skytemple_combo.setObjectName(u"skytemple_combo")

        self.goal_layout.addWidget(self.skytemple_combo)

        self.skytemple_slider_layout = QHBoxLayout()
        self.skytemple_slider_layout.setSpacing(6)
        self.skytemple_slider_layout.setObjectName(u"skytemple_slider_layout")
        self.skytemple_slider = QSlider(self.centralWidget)
        self.skytemple_slider.setObjectName(u"skytemple_slider")
        self.skytemple_slider.setMaximum(9)
        self.skytemple_slider.setPageStep(2)
        self.skytemple_slider.setOrientation(Qt.Horizontal)
        self.skytemple_slider.setTickPosition(QSlider.TicksBelow)

        self.skytemple_slider_layout.addWidget(self.skytemple_slider)

        self.skytemple_slider_label = QLabel(self.centralWidget)
        self.skytemple_slider_label.setObjectName(u"skytemple_slider_label")
        sizePolicy = QSizePolicy(QSizePolicy.Fixed, QSizePolicy.Fixed)
        sizePolicy.setHorizontalStretch(0)
        sizePolicy.setVerticalStretch(0)
        sizePolicy.setHeightForWidth(self.skytemple_slider_label.sizePolicy().hasHeightForWidth())
        self.skytemple_slider_label.setSizePolicy(sizePolicy)
        self.skytemple_slider_label.setMinimumSize(QSize(20, 0))
        self.skytemple_slider_label.setAlignment(Qt.AlignCenter)

        self.skytemple_slider_layout.addWidget(self.skytemple_slider_label)


        self.goal_layout.addLayout(self.skytemple_slider_layout)

        PresetEchoesGoal.setCentralWidget(self.centralWidget)

        self.retranslateUi(PresetEchoesGoal)

        QMetaObject.connectSlotsByName(PresetEchoesGoal)
    # setupUi

    def retranslateUi(self, PresetEchoesGoal):
        PresetEchoesGoal.setWindowTitle(QCoreApplication.translate("PresetEchoesGoal", u"Goal", None))
        self.skytemple_description.setText(QCoreApplication.translate("PresetEchoesGoal", u"<html><head/><body><p>Controls where the Sky Temple Keys will be located.</p><p>All Guardians and Sub-Guardians: One key will be placed in each of the guardians and sub-guardians.<br/>Guardians: One key will be placed as the reward of each of the guardians.<br/>Collect Sky Temple Keys: A configurable quantity will be shuffled over the game world.</p><p>The Guardians are: Amorbis, Chykka and Quadraxis.<br/>The Sub-Guardians are: Bomb Guardian, Jump Guardian, Boost Guardian, Grapple Guardian, Spider Guardian and Power Bomb Guardian.</p><p>You can always check Sky Temple Gateway for hints where the keys were placed.</p></body></html>", None))
        self.skytemple_combo.setItemText(0, QCoreApplication.translate("PresetEchoesGoal", u"Guardians and Sub-Guardians", None))
        self.skytemple_combo.setItemText(1, QCoreApplication.translate("PresetEchoesGoal", u"Guardians", None))
        self.skytemple_combo.setItemText(2, QCoreApplication.translate("PresetEchoesGoal", u"Collect Sky Temple Keys", None))

        self.skytemple_slider_label.setText(QCoreApplication.translate("PresetEchoesGoal", u"0", None))
    # retranslateUi




import typing
from pathlib import Path
from typing import Iterator, Optional

from PySide2.QtGui import QIcon
from PySide2.QtWidgets import QCheckBox, QApplication, QFileDialog, QMainWindow, QWidget, QComboBox, QTextEdit, \
    QLineEdit

from randovania import get_data_path
from randovania.layout.layout_description import LayoutDescription
from randovania.layout.preset_migration import VersionedPreset


def map_set_checked(iterable: Iterator[QCheckBox], new_status: bool):
    for checkbox in iterable:
        checkbox.setChecked(new_status)


def lock_application(value: bool):
    QApplication.instance().main_window.setEnabled(value)


def _prompt_user_for_file(window: QMainWindow,
                          caption: str,
                          filter: str,
                          dir: Optional[str] = None,
                          new_file: bool = False) -> Optional[Path]:
    """
    Helper function for all `prompt_user_for_*` functions.
    :param window:
    :param caption:
    :param filter:
    :param new_file: If false, prompt for an existing file.
    :return: A string if the user selected a file, None otherwise
    """
    if new_file:
        method = QFileDialog.getSaveFileName
    else:
        method = QFileDialog.getOpenFileName
    open_result = method(window, caption=caption, dir=dir, filter=filter)
    if not open_result or open_result == ("", ""):
        return None
    return Path(open_result[0])


def prompt_user_for_input_iso(window: QMainWindow) -> Optional[Path]:
    """
    Shows an QFileDialog asking the user for a vanilla Game ISO
    :param window:
    :return: A string if the user selected a file, None otherwise
    """
    return _prompt_user_for_file(window, caption="Select the vanilla Game ISO.", filter="*.iso")


def prompt_user_for_output_iso(window: QMainWindow,
                               default_name: str = "Randomized.iso") -> Optional[Path]:
    """
    Shows an QFileDialog asking the user where to place the output ISO
    :param window:
    :param default_name: Name of a file that will be offered by default in the UI.
    :return: A string if the user selected a file, None otherwise
    """
    return _prompt_user_for_file(window, caption="Where to place the Randomized Game ISO.",
                                 dir=default_name,
                                 filter="*.iso", new_file=True)


def prompt_user_for_output_game_log(window: QMainWindow, default_name: str) -> Optional[Path]:
    """
    Shows an QFileDialog asking the user for a Randovania seed log
    :param window:
    :param default_name:
    :return: A string if the user selected a file, None otherwise
    """
    return _prompt_user_for_file(window, caption="Select a Randovania seed log.",
                                 dir=default_name,
                                 filter="Randovania Game, *.{}".format(LayoutDescription.file_extension()),
                                 new_file=True)


def prompt_user_for_input_game_log(window: QMainWindow) -> Optional[Path]:
    """
    Shows an QFileDialog asking the user for a Randovania seed log
    :param window:
    :return: A string if the user selected a file, None otherwise
    """
    return _prompt_user_for_file(window, caption="Select a Randovania seed log.",
                                 filter="Randovania Game, *.{}".format(LayoutDescription.file_extension()),
                                 new_file=False)


def prompt_user_for_database_file(window: QMainWindow) -> Optional[Path]:
    """
    Shows an QFileDialog asking the user for a Randovania database file
    :param window:
    :return: A string if the user selected a file, None otherwise
    """
    return _prompt_user_for_file(window, caption="Select a Randovania database file.", filter="*.json")


def prompt_user_for_preset_file(window: QMainWindow, new_file: bool) -> Optional[Path]:
    """
    Shows an QFileDialog asking the user for a Randovania preset file
    :param window:
    :param new_file: If it should be an existing file (False) or not.
    :return: A path if the user selected a file, None otherwise
    """
    return _prompt_user_for_file(window, caption="Select a Randovania Preset file.",
                                 filter=f"Randovania Preset, *.{VersionedPreset.file_extension()};;"
                                        f"All Files (*.*)",
                                 new_file=new_file)


def set_default_window_icon(window: QWidget):
    """
    Sets the window icon for the given widget to the default icon
    :param window:
    :return:
    """
    window.setWindowIcon(QIcon(str(get_data_path().joinpath("icons", "sky_temple_key_NqN_icon.ico"))))


def set_combo_with_value(combo: QComboBox, value):
    """
    Searches all items of the given combo for the given value and changes the current index to that one.
    :param combo:
    :param value:
    :return:
    """
    combo.setCurrentIndex(combo.findData(value))


def set_error_border_stylesheet(edit: QTextEdit, has_error: bool):
    edit.has_error = has_error
    if has_error:
        edit.setStyleSheet(":enabled { border: 1px solid red; }"
                           ":disabled { border: 1px solid red; background: #CCC }")
    else:
        edit.setStyleSheet("")


def set_edit_if_different(edit: QLineEdit, new_text: str):
    """
    Sets the text of the given QLineEdit only if it differs from the current value.
    Prevents snapping the user's cursor to the end unnecessarily.
    :param edit:
    :param new_text:
    :return:
    """
    if edit.text() != new_text:
        edit.setText(new_text)


def get_network_client():
    from randovania.gui.lib.qt_network_client import QtNetworkClient
    return typing.cast(QtNetworkClient, QApplication.instance().network_client)


def get_game_connection():
    from randovania.game_connection.game_connection import GameConnection
    return typing.cast(GameConnection, QApplication.instance().game_connection)


def show_install_visual_cpp_redist():
    from PySide2 import QtWidgets

    download_url = 'https://aka.ms/vs/16/release/vc_redist.x64.exe'
    support_url = 'https://support.microsoft.com/en-us/help/2977003/the-latest-supported-visual-c-downloads'

    QtWidgets.QMessageBox.critical(
        None,
        "Unable to load Dolphin backend",
        "Please install the latest "
        f"<a href='{download_url}'>Microsoft Visual C++ Redistributable</a>.<br /><br />"
        f"For more details, see <a href='{support_url}'>Microsoft's webpage</a>.")



import json

from randovania.game_description import data_reader, data_writer
from randovania.games.prime import default_data


def test_round_trip_full():
    original_data = default_data.decode_default_prime2()

    game = data_reader.decode_data(original_data)
    encoded_data = data_writer.write_game_description(game)

    assert encoded_data == original_data


def test_round_trip_small(test_files_dir):
    # Setup
    with test_files_dir.joinpath("prime_data_as_json.json").open("r") as data_file:
        original_data = json.load(data_file)

    game = data_reader.decode_data(original_data)
    encoded_data = data_writer.write_game_description(game)

    assert encoded_data == original_data





import dataclasses
import functools

from PySide2 import QtWidgets
from PySide2.QtWidgets import QComboBox

from randovania.gui.generated.preset_echoes_beam_configuration_ui import Ui_PresetEchoesBeamConfiguration
from randovania.gui.preset_settings.preset_tab import PresetTab
from randovania.interface_common.preset_editor import PresetEditor
from randovania.layout.beam_configuration import BeamAmmoConfiguration
from randovania.layout.preset import Preset

_BEAMS = {
    "power": "Power Beam",
    "dark": "Dark Beam",
    "light": "Light Beam",
    "annihilator": "Annihilator Beam",
}


class PresetEchoesBeamConfiguration(PresetTab, Ui_PresetEchoesBeamConfiguration):
    _editor: PresetEditor

    def __init__(self, editor: PresetEditor):
        super().__init__()
        self.setupUi(self)
        self._editor = editor

        def _add_header(text: str, col: int):
            label = QtWidgets.QLabel(self.beam_configuration_group)
            label.setText(text)
            self.beam_configuration_layout.addWidget(label, 0, col)

        _add_header("Ammo A", 1)
        _add_header("Ammo B", 2)
        _add_header("Uncharged", 3)
        _add_header("Charged", 4)
        _add_header("Combo", 5)
        _add_header("Missiles for Combo", 6)

        self._beam_ammo_a = {}
        self._beam_ammo_b = {}
        self._beam_uncharged = {}
        self._beam_charged = {}
        self._beam_combo = {}
        self._beam_missile = {}

        def _create_ammo_combo():
            combo = QComboBox(self.beam_configuration_group)
            combo.addItem("None", -1)
            combo.addItem("Power Bomb", 43)
            combo.addItem("Missile", 44)
            combo.addItem("Dark Ammo", 45)
            combo.addItem("Light Ammo", 46)
            return combo

        row = 1
        for beam, beam_name in _BEAMS.items():
            label = QtWidgets.QLabel(self.beam_configuration_group)
            label.setText(beam_name)
            self.beam_configuration_layout.addWidget(label, row, 0)

            ammo_a = _create_ammo_combo()
            ammo_a.currentIndexChanged.connect(functools.partial(
                self._on_ammo_type_combo_changed, beam, ammo_a, False))
            self._beam_ammo_a[beam] = ammo_a
            self.beam_configuration_layout.addWidget(ammo_a, row, 1)

            ammo_b = _create_ammo_combo()
            ammo_b.currentIndexChanged.connect(functools.partial(
                self._on_ammo_type_combo_changed, beam, ammo_b, True))
            self._beam_ammo_b[beam] = ammo_b
            self.beam_configuration_layout.addWidget(ammo_b, row, 2)

            spin_box = QtWidgets.QSpinBox(self.beam_configuration_group)
            spin_box.setSuffix(" ammo")
            spin_box.setMaximum(250)
            spin_box.valueChanged.connect(functools.partial(
                self._on_ammo_cost_spin_changed, beam, "uncharged_cost"
            ))
            self._beam_uncharged[beam] = spin_box
            self.beam_configuration_layout.addWidget(spin_box, row, 3)

            spin_box = QtWidgets.QSpinBox(self.beam_configuration_group)
            spin_box.setSuffix(" ammo")
            spin_box.setMaximum(250)
            spin_box.valueChanged.connect(functools.partial(
                self._on_ammo_cost_spin_changed, beam, "charged_cost"
            ))
            self._beam_charged[beam] = spin_box
            self.beam_configuration_layout.addWidget(spin_box, row, 4)

            spin_box = QtWidgets.QSpinBox(self.beam_configuration_group)
            spin_box.setSuffix(" ammo")
            spin_box.setMaximum(250)
            spin_box.valueChanged.connect(functools.partial(
                self._on_ammo_cost_spin_changed, beam, "combo_ammo_cost"
            ))
            self._beam_combo[beam] = spin_box
            self.beam_configuration_layout.addWidget(spin_box, row, 5)

            spin_box = QtWidgets.QSpinBox(self.beam_configuration_group)
            spin_box.setSuffix(" missile")
            spin_box.setMaximum(250)
            spin_box.setMinimum(1)
            spin_box.valueChanged.connect(functools.partial(
                self._on_ammo_cost_spin_changed, beam, "combo_missile_cost"
            ))
            self._beam_missile[beam] = spin_box
            self.beam_configuration_layout.addWidget(spin_box, row, 6)

            row += 1

    @property
    def uses_patches_tab(self) -> bool:
        return True

    def _on_ammo_type_combo_changed(self, beam: str, combo: QComboBox, is_ammo_b: bool, _):
        with self._editor as editor:
            beam_configuration = editor.configuration.beam_configuration
            old_config: BeamAmmoConfiguration = getattr(beam_configuration, beam)
            if is_ammo_b:
                new_config = dataclasses.replace(old_config, ammo_b=combo.currentData())
            else:
                new_config = dataclasses.replace(old_config, ammo_a=combo.currentData())

            editor.set_configuration_field("beam_configuration",
                                           dataclasses.replace(beam_configuration, **{beam: new_config}))

    def _on_ammo_cost_spin_changed(self, beam: str, field_name: str, value: int):
        with self._editor as editor:
            beam_configuration = editor.configuration.beam_configuration
            new_config = dataclasses.replace(getattr(beam_configuration, beam),
                                             **{field_name: value})
            editor.set_configuration_field("beam_configuration",
                                           dataclasses.replace(beam_configuration, **{beam: new_config}))

    def on_preset_changed(self, preset: Preset):
        beam_configuration = preset.configuration.beam_configuration

        for beam in _BEAMS:
            config: BeamAmmoConfiguration = getattr(beam_configuration, beam)

            self._beam_ammo_a[beam].setCurrentIndex(self._beam_ammo_a[beam].findData(config.ammo_a))
            self._beam_ammo_b[beam].setCurrentIndex(self._beam_ammo_b[beam].findData(config.ammo_b))
            self._beam_ammo_b[beam].setEnabled(config.ammo_a != -1)
            self._beam_uncharged[beam].setValue(config.uncharged_cost)
            self._beam_charged[beam].setValue(config.charged_cost)
            self._beam_combo[beam].setValue(config.combo_ammo_cost)
            self._beam_missile[beam].setValue(config.combo_missile_cost)



from argparse import ArgumentParser


def flask_command_logic(args):
    from randovania.server import app
    server_app = app.create_app()
    server_app.sio.sio.run(server_app, host="0.0.0.0")


def add_flask_command(sub_parsers):
    parser: ArgumentParser = sub_parsers.add_parser(
        "flask",
        help="Hosts the flask server."
    )
    parser.set_defaults(func=flask_command_logic)


def bot_command_logic(args):
    from randovania.server import bot
    bot.run()


def add_bot_command(sub_parsers):
    parser: ArgumentParser = sub_parsers.add_parser(
        "bot",
        help="Runs the Discord bot."
    )
    parser.set_defaults(func=bot_command_logic)


def create_subparsers(sub_parsers):
    parser: ArgumentParser = sub_parsers.add_parser(
        "server",
        help="CLI tools for the server tools"
    )
    sub_parsers = parser.add_subparsers(dest="command")
    add_flask_command(sub_parsers)
    add_bot_command(sub_parsers)

    def check_command(args):
        if args.command is None:
            parser.print_help()
            raise SystemExit(1)

    parser.set_defaults(func=check_command)



from pathlib import Path


def patch_game_name_and_id(game_files_path: Path, new_name: str):
    b = new_name.encode("ASCII")
    if len(b) > 40:
        raise ValueError("Name '{}' is bigger than 40 bytes".format(new_name))

    with game_files_path.joinpath("sys", "boot.bin").open("r+b") as boot_bin:
        boot_bin.seek(0x5)
        boot_bin.write(b"R")
        boot_bin.seek(0x20)
        boot_bin.write(b)

    with game_files_path.joinpath("files", "opening.bnr").open("r+b") as banner:
        banner.seek(0x1860)
        banner.write(b)



from enum import Enum


class SessionAdminGlobalAction(Enum):
    """Actions that operate on the session itself"""
    CREATE_ROW = "create_row"
    CHANGE_ROW = "change_row"
    DELETE_ROW = "delete_row"
    UPDATE_LAYOUT_GENERATION = "update_layout_generation"
    CHANGE_LAYOUT_DESCRIPTION = "change_layout_description"
    DOWNLOAD_LAYOUT_DESCRIPTION = "download_layout_description"
    START_SESSION = "start_session"
    FINISH_SESSION = "finish_session"
    RESET_SESSION = "reset_session"
    CHANGE_PASSWORD = "change_password"
    DELETE_SESSION = "delete_session"


class SessionAdminUserAction(Enum):
    """Actions that operate on top of an user"""
    KICK = "kick"
    MOVE = "move"
    SWITCH_IS_OBSERVER = "switch_is_observer"
    SWITCH_ADMIN = "switch_admin"
    CREATE_PATCHER_FILE = "create_patcher_file"
    ABANDON = "abandon"



from typing import NamedTuple

from randovania.game_description.resources.resource_type import ResourceType


class SimpleResourceInfo(NamedTuple):
    index: int
    long_name: str
    short_name: str
    resource_type: ResourceType

    def __str__(self):
        return self.long_name



import dataclasses
from enum import Enum
from typing import List

from randovania.bitpacking.bitpacking import BitPackEnum, BitPackDataClass
from randovania.bitpacking.json_dataclass import JsonDataclass
from randovania.games.game import RandovaniaGame
from randovania.layout.base_configuration import BaseConfiguration
from randovania.layout.beam_configuration import BeamConfiguration
from randovania.layout.elevators import LayoutElevators
from randovania.layout.hint_configuration import HintConfiguration
from randovania.layout.translator_configuration import TranslatorConfiguration


class LayoutSkyTempleKeyMode(BitPackEnum, Enum):
    ALL_BOSSES = "all-bosses"
    ALL_GUARDIANS = "all-guardians"
    ZERO = 0
    ONE = 1
    TWO = 2
    THREE = 3
    FOUR = 4
    FIVE = 5
    SIX = 6
    SEVEN = 7
    EIGHT = 8
    NINE = 9

    @classmethod
    def default(cls) -> "LayoutSkyTempleKeyMode":
        return cls.NINE

    @property
    def num_keys(self):
        if self == self.ALL_BOSSES:
            return 9
        elif self == self.ALL_GUARDIANS:
            return 3
        else:
            return self.value


@dataclasses.dataclass(frozen=True)
class LayoutSafeZone(BitPackDataClass, JsonDataclass):
    fully_heal: bool
    prevents_dark_aether: bool
    heal_per_second: float = dataclasses.field(metadata={"min": 0.0, "max": 100.0,
                                                         "if_different": 1.0, "precision": 1.0})


@dataclasses.dataclass(frozen=True)
class EchoesConfiguration(BaseConfiguration):
    elevators: LayoutElevators
    sky_temple_keys: LayoutSkyTempleKeyMode
    translator_configuration: TranslatorConfiguration
    hints: HintConfiguration
    beam_configuration: BeamConfiguration
    skip_final_bosses: bool
    energy_per_tank: float = dataclasses.field(metadata={"min": 1.0, "max": 1000.0, "precision": 1.0})
    safe_zone: LayoutSafeZone
    menu_mod: bool
    warp_to_start: bool
    varia_suit_damage: float = dataclasses.field(metadata={"min": 0.1, "max": 60.0, "precision": 2.0})
    dark_suit_damage: float = dataclasses.field(metadata={"min": 0.0, "max": 60.0, "precision": 2.0})
    dangerous_energy_tank: bool

    @classmethod
    def game_enum(cls) -> RandovaniaGame:
        return RandovaniaGame.PRIME2

    def dangerous_settings(self) -> List[str]:
        result = super().dangerous_settings()

        if self.elevators == LayoutElevators.ONE_WAY_ANYTHING:
            result.append("One-way anywhere elevators")

        if self.dangerous_energy_tank:
            result.append("1 HP Mode")

        return result



import dataclasses
import struct
from typing import List

from randovania.dol_patching import assembler
from randovania.dol_patching.assembler import custom_ppc
from randovania.dol_patching.assembler.ppc import *
from randovania.dol_patching.dol_file import DolFile
from randovania.dol_patching.dol_version import DolVersion
from randovania.game_description.echoes_game_specific import EchoesGameSpecific
from randovania.games.game import RandovaniaGame


@dataclasses.dataclass(frozen=True)
class StringDisplayPatchAddresses:
    update_hint_state: int
    message_receiver_string_ref: int
    wstring_constructor: int
    display_hud_memo: int
    max_message_size: int


@dataclasses.dataclass(frozen=True)
class HealthCapacityAddresses:
    base_health_capacity: int
    energy_tank_capacity: int


@dataclasses.dataclass(frozen=True)
class DangerousEnergyTankAddresses:
    small_number_float: int
    incr_pickup: int


@dataclasses.dataclass(frozen=True)
class PowerupFunctionsAddresses:
    add_power_up: int
    incr_pickup: int
    decr_pickup: int


@dataclasses.dataclass(frozen=True)
class BasePrimeDolVersion(DolVersion):
    game_state_pointer: int
    cplayer_vtable: int
    cstate_manager_global: int
    string_display: StringDisplayPatchAddresses
    powerup_functions: PowerupFunctionsAddresses
    health_capacity: HealthCapacityAddresses
    game_options_constructor_address: int
    dangerous_energy_tank: DangerousEnergyTankAddresses


_registers_to_save = 2
_remote_execution_stack_size = 0x30 + (_registers_to_save * 4)
_remote_execution_max_byte_count = 296


def remote_execution_patch_start() -> List[BaseInstruction]:
    return_code = remote_execution_patch_end()
    intro = [
        # setup stack
        stwu(r1, -(_remote_execution_stack_size - 4), r1),
        mfspr(r0, LR),
        stw(r0, _remote_execution_stack_size, r1),
        stmw(GeneralRegister(32 - _registers_to_save), _remote_execution_stack_size - 4 - _registers_to_save * 4, r1),
        or_(r31, r3, r3),

        # return if displayed
        lbz(r4, 0x2, r31),
        cmpwi(r4, 0x0),
        bne((len(return_code) + 1) * 4, relative=True),

        # clean return if flag is not set
        *return_code,

        # set displayed
        li(r6, 0x0),
        stb(r6, 0x2, r31),
    ]

    num_bytes_to_invalidate = _remote_execution_max_byte_count - assembler.byte_count(intro)
    # Our loop end condition depends on this value being a multiple of 32, greater than 0
    num_bytes_to_invalidate = ((num_bytes_to_invalidate // 32) + 1) * 32

    return [
        *intro,

        # fetch the instructions again, since they're being overwritten externally
        # this clears Dolphin's JIT cache
        custom_ppc.load_current_address(r30, 8),
        li(r4, num_bytes_to_invalidate),

        icbi(4, 30),  # invalidate using r30 + r4
        cmpwi(r4, 0x0),
        addi(r4, r4, -32),
        bne(-3 * 4, relative=True),
        sync(),
        isync(),
    ]


def remote_execution_patch_end() -> List[BaseInstruction]:
    return [
        lmw(GeneralRegister(32 - _registers_to_save), _remote_execution_stack_size - 4 - _registers_to_save * 4, r1),
        lwz(r0, _remote_execution_stack_size, r1),
        mtspr(LR, r0),
        addi(r1, r1, _remote_execution_stack_size - 4),
        blr(),
    ]


def call_display_hud_patch(patch_addresses: StringDisplayPatchAddresses) -> List[BaseInstruction]:
    return [
        # setup CHUDMemoParms
        lis(r5, 0x4100),  # 8.0f
        li(r6, 0x0),
        li(r7, 0x1),
        li(r9, 0x9),
        stw(r5, 0x10, r1),  # display time (seconds)
        stb(r7, 0x14, r1),  # clear memo window
        stb(r6, 0x15, r1),  # fade out only
        stb(r6, 0x16, r1),  # hint memo
        stb(r7, 0x17, r1),  # fade in text
        stw(r9, 0x18, r1),  # unk

        # setup wstring
        addi(r3, r1, 0x1C),
        custom_ppc.load_unsigned_32bit(r4, patch_addresses.message_receiver_string_ref),
        bl(patch_addresses.wstring_constructor),

        # r4 = CHUDMemoParms
        addi(r4, r1, 0x10),
        bl(patch_addresses.display_hud_memo),
    ]


def adjust_item_amount_and_capacity_patch(patch_addresses: PowerupFunctionsAddresses, item_id: int, delta: int,
                                          ) -> List[BaseInstruction]:
    return [
        *increment_item_capacity_patch(patch_addresses, item_id, delta),

        lwz(r3, 0x150c, r31),
        li(r4, item_id),
        li(r5, abs(delta)),
        bl(patch_addresses.incr_pickup if delta >= 0 else patch_addresses.decr_pickup),
    ]


def increment_item_capacity_patch(patch_addresses: PowerupFunctionsAddresses, item_id: int, delta: int = 1
                                  ) -> List[BaseInstruction]:
    return [
        lwz(r3, 0x150c, r31),
        li(r4, item_id),
        li(r5, delta),
        bl(patch_addresses.add_power_up),
    ]


def apply_remote_execution_patch(patch_addresses: StringDisplayPatchAddresses, dol_file: DolFile):
    patch = [
        *remote_execution_patch_start(),
        *remote_execution_patch_end(),
    ]
    dol_file.write_instructions(patch_addresses.update_hint_state, patch)


def create_remote_execution_body(patch_addresses: StringDisplayPatchAddresses,
                                 instructions: List[BaseInstruction]) -> Tuple[int, bytes]:
    """
    Return the address and the bytes for executing the given instructions via remote code execution.
    """
    update_hint_state = patch_addresses.update_hint_state

    remote_start_instructions = remote_execution_patch_start()
    remote_start_byte_count = assembler.byte_count(remote_start_instructions)

    body_address = update_hint_state + remote_start_byte_count
    body_instructions = list(instructions)
    body_instructions.extend(remote_execution_patch_end())
    body_bytes = bytes(assembler.assemble_instructions(body_address, body_instructions))

    if len(body_bytes) > _remote_execution_max_byte_count - remote_start_byte_count:
        raise ValueError(f"Received {len(body_instructions)} instructions with total {len(body_bytes)} bytes, "
                         f"but limit is {_remote_execution_max_byte_count - remote_start_byte_count}.")

    return body_address, body_bytes


def apply_energy_tank_capacity_patch(patch_addresses: HealthCapacityAddresses, game_specific: EchoesGameSpecific,
                                     dol_file: DolFile):
    """
    Patches the base health capacity and the energy tank capacity with matching values.
    """
    tank_capacity = game_specific.energy_per_tank

    dol_file.write(patch_addresses.base_health_capacity, struct.pack(">f", tank_capacity - 1))
    dol_file.write(patch_addresses.energy_tank_capacity, struct.pack(">f", tank_capacity))


def apply_reverse_energy_tank_heal_patch(sd2_base: int,
                                         addresses: DangerousEnergyTankAddresses,
                                         active: bool,
                                         game: RandovaniaGame,
                                         dol_file: DolFile,
                                         ):
    if game == RandovaniaGame.PRIME2:
        health_offset = 0x14
        refill_item = 0x29
        patch_offset = 0x90

    elif game == RandovaniaGame.PRIME3:
        health_offset = 0xc
        refill_item = 0x12
        patch_offset = 0x138

    else:
        raise ValueError(f"Unsupported game: {game}")

    if active:
        patch = [
            lfs(f0, (addresses.small_number_float - sd2_base), r2),
            stfs(f0, health_offset, r30),
            ori(r0, r0, 0),
            ori(r0, r0, 0),
        ]
    else:
        patch = [
            or_(r3, r30, r30),
            li(r4, refill_item),
            li(r5, 9999),
            bl(addresses.incr_pickup),
        ]

    dol_file.write_instructions(addresses.incr_pickup + patch_offset, patch)
